# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import pydantic

import spectre_core.events
import spectre_core.flowgraphs

from ._validators import (
    skip_validator,
    validate_window_size,
    validate_nyquist_criterion,
    validate_non_overlapping_steps,
    validate_num_steps_per_sweep,
    validate_num_samples_per_step,
)
from ._usrp_validators import (
    validate_wire_format,
    validate_sample_rate_with_master_clock_rate,
)


class USRPFixedCenterFrequency(
    spectre_core.flowgraphs.USRPFixedCenterFrequencyModel,
    spectre_core.events.FixedCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_nyquist_criterion(self.sample_rate, self.bandwidth)
        validate_window_size(self.window_size)
        validate_wire_format(self.wire_format)
        validate_sample_rate_with_master_clock_rate(
            self.sample_rate, self.master_clock_rate
        )
        return self


class USRPSweptCenterFrequency(
    spectre_core.flowgraphs.USRPSweptCenterFrequencyModel,
    spectre_core.events.SweptCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_window_size(self.window_size)
        validate_non_overlapping_steps(self.frequency_step, self.sample_rate)
        validate_num_samples_per_step(self.window_size, self.samples_per_step)
        validate_num_steps_per_sweep(
            self.min_frequency, self.max_frequency, self.frequency_step
        )
        validate_sample_rate_with_master_clock_rate(
            self.sample_rate, self.master_clock_rate
        )
        return self



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import typing
import datetime

import numpy as np
import numpy.typing as npt
import pyfftw

import spectre_core.batches
import spectre_core.exceptions
import spectre_core.spectrograms
import spectre_core.fields

from ._base import Base, BaseModel
from ._stfft import (
    get_buffer,
    get_window,
    get_fftw_obj,
    get_frequencies,
    get_num_spectrums,
    stfft,
    WindowType,
)

_LOGGER = logging.getLogger(__name__)


def _stitch_steps(
    stepped_dynamic_spectra: npt.NDArray[np.float32], num_full_sweeps: int
) -> npt.NDArray[np.float32]:
    """For each full sweep, create a single spectrum by stitching together the spectrum at each step."""
    return stepped_dynamic_spectra.reshape((num_full_sweeps, -1)).T


def _average_over_steps(
    stepped_dynamic_spectra: npt.NDArray[np.float32],
) -> npt.NDArray[np.float32]:
    """Average the spectrums in each step totally in time."""
    return np.nanmean(stepped_dynamic_spectra[..., 1:], axis=-1)


def _compute_num_max_frames_in_step(
    num_samples: npt.NDArray[np.int32], window_size: int, window_hop: int
) -> int:
    """Compute the maximum number of frames for all steps, and all sweeps, in the batch."""
    return get_num_spectrums(int(np.max(num_samples)), window_size, window_hop)


def _compute_num_full_sweeps(center_frequencies: npt.NDArray[np.float32]) -> int:
    """Compute the total number of full sweeps in the batch.

    Since the number of each samples in each step is variable, we only know a sweep is complete
    when there is a sweep after it. So we can define the total number of *full* sweeps as the number of
    (freq_max, freq_min) pairs in center_frequencies. It is only at an instance of (freq_max, freq_min) pair
    in center frequencies that the frequency decreases, so, we can compute the number of full sweeps by
    counting the numbers of negative values in np.diff(center_frequencies).
    """
    return len(np.where(np.diff(center_frequencies) < 0)[0])


def _compute_num_steps_per_sweep(center_frequencies: npt.NDArray[np.float32]) -> int:
    """Compute the (ensured constant) number of steps in each sweep."""
    # Find the (step) indices corresponding to the minimum frequencies.
    min_freq_indices = np.where(center_frequencies == np.min(center_frequencies))[0]
    # Then, we evaluate the number of steps that has occured between them via np.diff over the indices.
    unique_num_steps_per_sweep = np.unique(np.diff(min_freq_indices))
    # We expect that the difference is always the same, so that the result of np.unique has a single element.
    if len(unique_num_steps_per_sweep) != 1:
        raise spectre_core.exceptions.InvalidSweepMetadataError(
            (
                "Irregular step count per sweep, "
                "expected a consistent number of steps per sweep"
            )
        )
    return int(unique_num_steps_per_sweep[0])


def _validate_center_frequencies_ordering(
    center_frequencies: npt.NDArray[np.float32], freq_step: float
) -> None:
    """Check that the center frequencies are well-ordered in the detached header."""
    min_frequency = np.min(center_frequencies)
    # Extract the expected difference between each step within a sweep.
    for i, diff in enumerate(np.diff(center_frequencies)):
        # The steps should either increase by freq_step or drop to the minimum.
        if (diff != freq_step) and (center_frequencies[i + 1] != min_frequency):
            raise spectre_core.exceptions.InvalidSweepMetadataError(
                f"Unordered center frequencies detected, I/Q samples have been mixed up"
            )


def _compute_stepped_dynamic_spectra(
    stepped_dynamic_spectra: npt.NDArray[np.float32],
    fftw_obj: pyfftw.FFTW,
    buffer: npt.NDArray[np.complex64],
    iq_data: npt.NDArray[np.complex64],
    window: npt.NDArray[np.float32],
    window_hop: int,
    num_full_sweeps: int,
    num_steps_per_sweep: int,
    num_samples: npt.NDArray[np.int32],
) -> None:
    """For each full sweep, compute execute a short-time discrete Fourier transform on the IQ samples for each step."""
    # Store the step index over all sweeps (doesn't reset each sweep).
    global_step_index = 0
    # Store the index of the first sample in the step.
    start_sample_index = 0

    for sweep_index in range(num_full_sweeps):
        for step_index in range(num_steps_per_sweep):

            # Extract how many samples are in the current step from the metadata.
            end_sample_index = start_sample_index + num_samples[global_step_index]

            # Compute the number of frames we can squeeze into the current step.
            num_frames = get_num_spectrums(
                num_samples[global_step_index], window.size, window_hop
            )

            # Execute a short time discrete fourier transform on the step, then shift the
            # zero-frequency component to the middle of the spectrum.
            stepped_dynamic_spectra[sweep_index, step_index, :, :num_frames] = (
                np.fft.fftshift(
                    stfft(
                        fftw_obj,
                        buffer,
                        iq_data[start_sample_index:end_sample_index],
                        window,
                        window_hop,
                    ),
                    axes=0,
                )
            )

            # Reassign the start_sample_index for the next step.
            start_sample_index = end_sample_index

            # Finally, increment the global step index
            global_step_index += 1


def _compute_frequencies(
    frequencies: npt.NDArray[np.float32],
    center_frequencies: npt.NDArray[np.float32],
    window_size: int,
    sample_rate: int,
) -> None:
    """Assign physical frequencies to each of the spectral components in the stitched spectrum."""
    baseband_frequencies = np.fft.fftshift(get_frequencies(window_size, sample_rate))

    for i, center_frequency in enumerate(np.unique(center_frequencies)):
        lower_bound = i * window_size
        upper_bound = (i + 1) * window_size
        frequencies[lower_bound:upper_bound] = baseband_frequencies + center_frequency


def _compute_times(
    times: npt.NDArray[np.float32],
    num_samples: npt.NDArray[np.int32],
    sample_rate: float,
    num_full_sweeps: int,
    num_steps_per_sweep: int,
) -> None:
    """Assign physical times to each swept spectrum. We use (by convention) the time of the first sample in each sweep"""
    sampling_interval = 1 / sample_rate
    cumulative_samples = 0
    for sweep_index in range(num_full_sweeps):
        # Assign a physical time to the spectrum for this sweep
        times[sweep_index] = cumulative_samples * sampling_interval

        # Find the total number of samples across the sweep.
        start_step = sweep_index * num_steps_per_sweep
        end_step = (sweep_index + 1) * num_steps_per_sweep

        # Update cumulative samples
        cumulative_samples += np.sum(num_samples[start_step:end_step])


def _swept_stfft(
    fftw_obj: pyfftw.FFTW,
    buffer: npt.NDArray[np.complex64],
    iq_data: npt.NDArray[np.complex64],
    window: npt.NDArray[np.float32],
    window_hop: int,
    sample_rate: int,
    frequency_step: float,
    center_frequencies: npt.NDArray[np.float32],
    num_samples: npt.NDArray[np.int32],
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.float32], npt.NDArray[np.float32]]:
    _validate_center_frequencies_ordering(center_frequencies, frequency_step)

    num_steps_per_sweep = _compute_num_steps_per_sweep(center_frequencies)
    num_full_sweeps = _compute_num_full_sweeps(center_frequencies)
    num_max_frames_in_step = _compute_num_max_frames_in_step(
        num_samples, window.size, window_hop
    )

    stepped_dynamic_spectra_shape = (
        num_full_sweeps,
        num_steps_per_sweep,
        window.size,
        num_max_frames_in_step,
    )

    # Pad with nan values up to the max number of frames to make computations simpler.
    stepped_dynamic_spectra = np.full(
        stepped_dynamic_spectra_shape, np.nan, dtype=np.float32
    )
    _compute_stepped_dynamic_spectra(
        stepped_dynamic_spectra,
        fftw_obj,
        buffer,
        iq_data,
        window,
        window_hop,
        num_full_sweeps,
        num_steps_per_sweep,
        num_samples,
    )

    # Assign physical frequencies to each spectral component in the spectrum for each sweep.
    frequencies = np.empty(num_steps_per_sweep * window.size, dtype=np.float32)
    _compute_frequencies(frequencies, center_frequencies, window.size, sample_rate)

    # Assign physical times to each spectrum in the spectrogram.
    times = np.empty(num_full_sweeps, dtype=np.float32)
    _compute_times(
        times, num_samples, sample_rate, num_full_sweeps, num_steps_per_sweep
    )

    dynamic_spectra = _stitch_steps(
        _average_over_steps(stepped_dynamic_spectra), num_full_sweeps
    )

    return times, frequencies, dynamic_spectra


def _prepend_num_samples(
    carryover_num_samples: npt.NDArray[np.int32],
    num_samples: npt.NDArray[np.int32],
    final_step_spans_two_batches: bool,
) -> npt.NDArray[np.int32]:
    """Prepend the number of samples from the final sweep of the previous batch, to the first
    sweep of the current batch."""
    if final_step_spans_two_batches:
        # In the case that the step has bled across batches, adjust the number of samples accordingly.
        num_samples[0] += carryover_num_samples[-1]
        return np.concatenate((carryover_num_samples[:-1], num_samples))
    else:
        return np.concatenate((carryover_num_samples, num_samples))


def _prepend_center_frequencies(
    carryover_center_frequencies: npt.NDArray[np.float32],
    center_frequencies: npt.NDArray[np.float32],
    final_step_spans_two_batches: bool,
) -> npt.NDArray[np.float32]:
    """Prepend the center frequencies from the final sweep of the previous batch, to the first
    sweep of the current batch."""
    if final_step_spans_two_batches:
        # In the case that final step has bled across batches, do not permit identical neighbours in the center frequency array.
        return np.concatenate((carryover_center_frequencies[:-1], center_frequencies))
    else:
        return np.concatenate((carryover_center_frequencies, center_frequencies))


def _prepend_iq_data(
    carryover_iq_data: npt.NDArray[np.complex64], iq_data: npt.NDArray[np.complex64]
) -> npt.NDArray[np.complex64]:
    """Prepend the IQ samples from the final sweep of the previous batch, to the first sweep
    of the current batch."""
    return np.concatenate((carryover_iq_data, iq_data))


def _get_final_sweep(
    previous_iq_data: npt.NDArray[np.complex64],
    previous_iq_metadata: spectre_core.batches.IQMetadata,
) -> tuple[npt.NDArray[np.complex64], npt.NDArray[np.float32], npt.NDArray[np.int32]]:
    """Get IQ samples and metadata from the final sweep of the previous batch."""

    if (
        previous_iq_metadata.center_frequencies is None
        or previous_iq_metadata.num_samples is None
    ):
        raise ValueError(f"Expected non-empty IQ metadata.")

    # Find the step index from the last sweep
    # [0] since the return of np.where is a 1 element Tuple,
    # containing a list of step indices corresponding to the smallest center frequencies
    # [-1] since we want the final step index, where the center frequency is minimised
    final_sweep_start_step_index = np.where(
        previous_iq_metadata.center_frequencies
        == np.min(previous_iq_metadata.center_frequencies)
    )[0][-1]

    # Isolate the data from the final sweep.
    final_center_frequencies = previous_iq_metadata.center_frequencies[
        final_sweep_start_step_index:
    ]
    final_num_samples = previous_iq_metadata.num_samples[final_sweep_start_step_index:]
    final_iq_data = previous_iq_data[-np.sum(final_num_samples) :]

    # Do a sanity check on the number of samples in the final sweep
    if len(final_iq_data) != np.sum(final_num_samples):
        raise ValueError(
            (
                f"Unexpected error! Mismatch in sample count for the final sweep data."
                f"Expected {np.sum(final_num_samples)} based on sweep metadata, but found "
                f" {len(final_iq_data)} IQ samples in the final sweep"
            )
        )

    return final_iq_data, final_center_frequencies, final_num_samples


def _reconstruct_initial_sweep(
    previous_iq_data: npt.NDArray[np.complex64],
    previous_iq_metadata: spectre_core.batches.IQMetadata,
    iq_data: npt.NDArray[np.complex64],
    iq_metadata: spectre_core.batches.IQMetadata,
) -> tuple[
    npt.NDArray[np.complex64], npt.NDArray[np.float32], npt.NDArray[np.int32], int
]:
    """Reconstruct the initial sweep of the current batch, using data from the previous batch.

    Specifically, we extract the data from the final sweep of the previous batch and prepend
    it to the first sweep of the current batch. Additionally, we return how many IQ samples
    we prepended, which will allow us to correct the spectrogram start time of the current batch.
    """

    if iq_metadata.center_frequencies is None or iq_metadata.num_samples is None:
        raise ValueError(f"Expected non-empty IQ metadata")

    # Extract the final sweep of the previous batch, to carryover to the first sweep of the current batch.
    carryover_iq_data, carryover_center_frequencies, carryover_num_samples = (
        _get_final_sweep(previous_iq_data, previous_iq_metadata)
    )

    # Prepend the iq data from the final sweep of the previous batch.
    iq_data = _prepend_iq_data(carryover_iq_data, iq_data)

    # Assess whether the final step in the previous batch, bleeds to the next.
    final_step_spans_two_batches = (
        carryover_center_frequencies[-1] == iq_metadata.center_frequencies[0]
    )

    # Prepend the iq metadata from the final sweep of the previous batch.
    center_frequencies = _prepend_center_frequencies(
        carryover_center_frequencies,
        iq_metadata.center_frequencies,
        final_step_spans_two_batches,
    )
    num_samples = _prepend_num_samples(
        carryover_num_samples, iq_metadata.num_samples, final_step_spans_two_batches
    )

    # Keep track of how many samples we prepended, so we can adjust the timing later.
    num_samples_prepended = int(np.sum(carryover_num_samples))

    return (iq_data, center_frequencies, num_samples, num_samples_prepended)


class SweptCenterFrequencyModel(BaseModel):
    window_size: spectre_core.fields.Field.window_size = 1024
    window_hop: spectre_core.fields.Field.window_hop = 1024
    window_type: spectre_core.fields.Field.window_type = "blackman"
    sample_rate: spectre_core.fields.Field.sample_rate = 1000000
    frequency_resolution: spectre_core.fields.Field.frequency_resolution = 0
    time_resolution: spectre_core.fields.Field.time_resolution = 0
    keep_signal: spectre_core.fields.Field.keep_signal = False
    frequency_step: spectre_core.fields.Field.frequency_step = 1000000


class SweptCenterFrequency(
    Base[SweptCenterFrequencyModel, spectre_core.batches.IQStreamBatch]
):
    def __init__(
        self,
        tag: str,
        model: SweptCenterFrequencyModel,
        batch_cls: typing.Type[spectre_core.batches.IQStreamBatch],
    ) -> None:
        super().__init__(tag, model, batch_cls)
        self.__model = model
        self.__window = get_window(
            WindowType(self.__model.window_type), self.__model.window_size
        )

        # Pre-allocate the buffer.
        self.__buffer = get_buffer(self.__model.window_size)

        # Defer the expensive FFTW plan creation until the first batch is being processed.
        # With this approach, we avoid a bug where filesystem events are missed because
        # the watchdog observer isn't set up in time before the receiver starts capturing data.
        self.__fftw_obj = None

        # Initialise a cache to hold the previous batches data.
        self.__previous_batch: typing.Optional[spectre_core.batches.IQStreamBatch] = (
            None
        )

    @property
    def _watch_extension(self) -> str:
        return spectre_core.batches.IQStreamBatchExtension.BIN

    def process(
        self, batch: spectre_core.batches.IQStreamBatch
    ) -> spectre_core.spectrograms.Spectrogram:
        """Compute a spectrogram of IQ samples captured at a center frequency periodically swept in
        fixed increments. Neighbouring IQ samples captured at the same frequency constitute a "step". Neighbouring
        steps captured at incrementally increasing center frequencies form a "sweep." A new sweep begins when the
        center frequency resets to its minimum value.

        A short-time discrete Fourier transform is computed for each step, with the resulting spectra averaged over
        in time to produce a single spectrum per step. The spectra for each step are stitched in frequency to form a single
        spectrum for each sweep. These swept spectra are stiched in time to produce the final spectrogram, which is saved to file
        in the FITS format.
        """
        _LOGGER.info(f"Reading {batch.bin_file.file_name}")
        iq_data = batch.bin_file.cached_read()

        _LOGGER.info(f"Reading {batch.hdr_file.file_name}")
        iq_metadata = batch.hdr_file.cached_read()

        center_frequencies, num_samples = (
            iq_metadata.center_frequencies,
            iq_metadata.num_samples,
        )
        if center_frequencies is None or num_samples is None:
            raise ValueError(
                f"An unexpected error has occured, expected frequency tag metadata "
                f"in the detached header."
            )

        # Account for the millisecond correction.
        start_datetime = batch.start_datetime + datetime.timedelta(
            milliseconds=iq_metadata.millisecond_correction
        )

        # If a previous batch is stored, the initial sweep may span two adjacent batched files.
        if self.__previous_batch is not None:
            # If this is the case, first reconstruct the initial sweep of the current batch
            # by prepending the final sweep of the previous batch.
            iq_data, center_frequencies, num_samples, num_samples_prepended = (
                _reconstruct_initial_sweep(
                    self.__previous_batch.bin_file.cached_read(),
                    self.__previous_batch.hdr_file.cached_read(),
                    iq_data,
                    iq_metadata,
                )
            )

            # Since we have prepended extra samples, we need to correct the spectrogram start time appropriately.
            elapsed_time = num_samples_prepended * (1 / self.__model.sample_rate)
            start_datetime -= datetime.timedelta(seconds=float(elapsed_time))

        if self.__fftw_obj is None:
            _LOGGER.info(f"Creating the FFTW plan")
            self.__fftw_obj = get_fftw_obj(self.__buffer)

        _LOGGER.info("Executing the short-time FFT")
        # Compute the short-time discrete fourier transform.
        times, frequencies, dynamic_spectra = _swept_stfft(
            self.__fftw_obj,
            self.__buffer,
            iq_data,
            self.__window,
            self.__model.window_hop,
            self.__model.sample_rate,
            self.__model.frequency_step,
            center_frequencies,
            num_samples,
        )

        _LOGGER.info("Creating the spectrogram")
        spectrogram = spectre_core.spectrograms.Spectrogram(
            dynamic_spectra,
            times,
            frequencies,
            spectre_core.spectrograms.SpectrumUnit.AMPLITUDE,
            start_datetime,
        )

        spectrogram = spectre_core.spectrograms.time_average(
            spectrogram, resolution=self.__model.time_resolution
        )
        spectrogram = spectre_core.spectrograms.frequency_average(
            spectrogram, resolution=self.__model.frequency_resolution
        )

        # If the previous batch exists, then by this point it has already been processed.
        if self.__previous_batch is not None:
            if not self.__model.keep_signal:
                _LOGGER.info(f"Deleting {self.__previous_batch.bin_file.file_path}")
                self.__previous_batch.bin_file.delete()

                _LOGGER.info(f"Deleting {self.__previous_batch.hdr_file.file_path}")
                self.__previous_batch.hdr_file.delete()

        # Assign the current batch to be used as the previous batch at the next call of this method.
        self.__previous_batch = batch

        return spectrogram



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Real-time, extensible post-processing of SDR data into spectrograms."""

from ._base import Base
from ._fixed_center_frequency import FixedCenterFrequency, FixedCenterFrequencyModel
from ._swept_center_frequency import SweptCenterFrequency, SweptCenterFrequencyModel

from ._stfft import (
    stfft,
    get_buffer,
    get_window,
    get_fftw_obj,
    get_frequencies,
    get_times,
    get_cosine_signal,
    get_num_spectrums,
    WindowType,
)

__all__ = [
    "Base",
    "FixedCenterFrequency",
    "FixedCenterFrequencyModel",
    "SweptCenterFrequency",
    "SweptCenterFrequencyModel",
    "stfft",
    "get_buffer",
    "get_window",
    "get_fftw_obj",
    "get_times",
    "get_frequencies",
    "get_num_spectrums",
    "get_cosine_signal",
    "WindowType",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum


class ProcessType(Enum):
    """The origin of a Spectre process.

    USER: The main user session.
    WORKER: A background task created and managed internally by Spectre.
    """

    USER = "user"
    WORKER = "worker"



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from typing import Callable, TypeVar, ParamSpec
from functools import wraps

# ParamSpec for capturing the argument types of the function
P = ParamSpec("P")
# TypeVar for capturing the return type of the function
RT = TypeVar("RT")


def log_call(func: Callable[P, RT]) -> Callable[P, RT]:
    """Decorator to log the execution of a function.

    Logs an informational message when the decorated function is called,
    and an error message if the function raises an exception.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> RT:
        logger = logging.getLogger(func.__module__)
        try:
            logger.info(f"Calling the function: {func.__name__}")
            return func(*args, **kwargs)
        except Exception as e:
            logger.error(f"Error in function: {func.__name__}", exc_info=True)
            raise

    return wrapper



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing
import logging

from ._validators import validate_in_range, validate_one_of

_LOGGER = logging.getLogger(__name__)

SAMPLE_RATE_LOWER_BOUND = 62.5e3
SAMPLE_RATE_UPPER_BOUND = 10.66e6
CENTER_FREQUENCY_LOWER_BOUND = 1e3
CENTER_FREQUENCY_UPPER_BOUND = 2e9
IF_GAIN_UPPER_BOUND = -20
IF_GAIN_LOWER_BOUND = -59
RF_GAIN_UPPER_BOUND = 0
API_RETUNING_LATENCY = 25 * 1e-3
LOW_IF_SAMPLE_RATE_CUTOFF = 2e6
LOW_IF_PERMITTED_SAMPLE_RATES = [LOW_IF_SAMPLE_RATE_CUTOFF / (2**i) for i in range(6)]
# bandwidth == 0 means 'AUTO', i.e. the largest bandwidth compatible with the sample rate
BANDWIDTH_OPTIONS = [0, 200e3, 300e3, 600e3, 1.536e6, 5e6, 6e6, 7e6, 8e6]
LOW_IF_SAMPLE_RATE_CUTOFF = 2e6
LOW_IF_PERMITTED_SAMPLE_RATES = [LOW_IF_SAMPLE_RATE_CUTOFF / (2**i) for i in range(6)]


def validate_center_frequency(center_frequency: float) -> None:
    validate_in_range(
        center_frequency,
        lower_bound=CENTER_FREQUENCY_LOWER_BOUND,
        upper_bound=CENTER_FREQUENCY_UPPER_BOUND,
        name="center_frequency",
    )


def validate_constant_lna_state(
    min_frequency: float,
    max_frequency: float,
    get_rf_gains: typing.Callable[[float], list[int]],
):
    if get_rf_gains(min_frequency) != get_rf_gains(max_frequency):
        _LOGGER.warning(
            "Crossing a threshold where the LNA state has to change. Performance may be reduced."
        )


def validate_sample_rate(sample_rate: int) -> None:
    validate_in_range(
        sample_rate,
        lower_bound=SAMPLE_RATE_LOWER_BOUND,
        upper_bound=SAMPLE_RATE_UPPER_BOUND,
        name="sample_rate",
    )


def validate_bandwidth(bandwidth: float) -> None:
    validate_one_of(bandwidth, BANDWIDTH_OPTIONS, name="bandwidth")


def validate_if_gain(if_gain: float) -> None:
    validate_in_range(
        if_gain,
        lower_bound=IF_GAIN_LOWER_BOUND,
        upper_bound=IF_GAIN_UPPER_BOUND,
        name="if_gain",
    )


def validate_low_if_sample_rate(sample_rate: int) -> None:
    """Validate the sample rate if the receiver is operating in low IF mode.

    The minimum physical sampling rate of the SDRplay hardware is 2 MHz. Lower effective rates can be achieved
    through decimation, as handled by the `gr-sdrplay3` OOT module. This function ensures that the sample rate
    is not silently adjusted by the backend.

    For implementation details, refer to:
    https://github.com/fventuri/gr-sdrplay3/blob/v3.11.0.9/lib/rsp_impl.cc#L140-L179
    """

    if sample_rate <= LOW_IF_SAMPLE_RATE_CUTOFF:
        if sample_rate not in LOW_IF_PERMITTED_SAMPLE_RATES:
            raise ValueError(
                f"If the requested sample rate is less than or equal to {LOW_IF_SAMPLE_RATE_CUTOFF}, "
                f"the receiver will be operating in low IF mode. "
                f"So, the sample rate must be exactly one of {LOW_IF_PERMITTED_SAMPLE_RATES}. "
                f"Got sample rate {sample_rate} Hz"
            )


def validate_rf_gain(rf_gain: float, expected_rf_gains: list[int]):
    """Validate the RF gain value against the expected values for the current LNA state.

    The RF gain is determined by the LNA state and can only take specific values as documented in the
    gain reduction tables of the SDRplay API specification.

    For implementation details, refer to the `gr-sdrplay3` OOT module:
    https://github.com/fventuri/gr-sdrplay3/blob/v3.11.0.9/lib/rsp_impl.cc#L378-L387
    """
    if rf_gain not in expected_rf_gains:
        raise ValueError(
            f"The value of RF gain must be one of {expected_rf_gains}. "
            f"Got {rf_gain}."
        )



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing
import logging
import pydantic

import watchdog.observers
import watchdog.events

import spectre_core.exceptions
import spectre_core.batches
import spectre_core.events
import spectre_core.config
import spectre_core.flowgraphs
import spectre_core.logs

from ._config import Config, read_config, write_config

_LOGGER = logging.getLogger(__name__)

T = typing.TypeVar("T")


class ReceiverComponents(typing.Generic[T]):
    def __init__(self) -> None:
        """Manage receiver components per operating mode."""
        self._components: dict[str, T] = {}

    @property
    def modes(self) -> list[str]:
        """Get all the added operating modes."""
        return list(self._components.keys())

    def add(self, mode: str, component: T) -> None:
        """Add a component for a particular operating mode.

        :param mode: The operating mode for the receiver.
        :param component: The component associated with this mode.
        """
        self._components[mode] = component

    def get(self, mode: str) -> T:
        """Retrieve the component for a particular operating mode.

        :param mode: The operating mode for the receiver.
        :return: The component associated with this mode.
        :raises ModeNotFoundError: If the mode is not found.
        """
        if mode not in self._components:
            raise spectre_core.exceptions.ModeNotFoundError(
                f"Mode `{mode}` not found. Expected one of {self.modes}"
            )
        return self._components[mode]


class Models(ReceiverComponents[typing.Type[pydantic.BaseModel]]):
    """Store the model class per operating mode."""


class Flowgraphs(ReceiverComponents[typing.Type[spectre_core.flowgraphs.Base]]):
    """Store the flowgraph class per operating mode."""


class EventHandlers(ReceiverComponents[typing.Type[spectre_core.events.Base]]):
    """Store the event handler class per operating mode."""


class Batches(ReceiverComponents[typing.Type[spectre_core.batches.Base]]):
    """Store the batch class per operating mode."""


class Base:
    def __init__(
        self,
        name: str,
        mode: typing.Optional[str] = None,
        models: typing.Optional[Models] = None,
        flowgraphs: typing.Optional[Flowgraphs] = None,
        event_handlers: typing.Optional[EventHandlers] = None,
        batches: typing.Optional[Batches] = None,
    ) -> None:
        """An abstraction layer for software-defined radio receivers.

        :param name: The name of the receiver.
        :param mode: Optionally set the operating mode for the receiver, defaults to None
        :param models: Optionally provide model classes per operating mode, defaults to None
        :param flowgraphs: Optionally provide flowgraph classes per operating mode, defaults to None
        :param event_handlers: Optionally provide event handler classes per operating mode, defaults to None
        :param batches: Optionally provide batch classes per operating mode, defaults to None
        """
        self.__name = name
        self.__mode = mode
        self.__models = models or Models()
        self.__flowgraphs = flowgraphs or Flowgraphs()
        self.__event_handlers = event_handlers or EventHandlers()
        self.__batches = batches or Batches()

    @property
    def name(self) -> str:
        """The name of the receiver."""
        return self.__name

    @property
    def mode(self) -> typing.Optional[str]:
        """The operating mode, or `None` if it's not set."""
        return self.__mode

    @mode.setter
    def mode(self, value: typing.Optional[str]) -> None:
        """Set the operating mode.

        :param value: The new operating mode of the receiver. Use `None` to unset the mode.
        """
        _LOGGER.info(f"Setting the mode to '{value}'")
        if not value is None and value not in self.modes:
            raise spectre_core.exceptions.ModeNotFoundError(
                f"Mode `{value}` not found. Expected one of {self.modes}"
            )
        self.__mode = value

    @property
    def modes(self) -> list[str]:
        """The available operating modes for the receiver.

        :raises ValueError: If the modes are inconsistent.
        """
        if (
            not self.__flowgraphs.modes
            == self.__event_handlers.modes
            == self.__batches.modes
        ):
            raise ValueError(f"Inconsistent modes for the receiver '{self.name}'")
        return self.__flowgraphs.modes

    @property
    def active_mode(self) -> str:
        """The active operating mode, raising an error if not set.

        :raises ValueError: If no mode is currently set.
        :return: The active operating mode.
        """
        if self.__mode is None:
            raise ValueError(
                f"An active mode is not set for the receiver '{self.name}'. "
                f"Currently, the mode is {self.__mode}"
            )
        return self.__mode

    @property
    def model_cls(self) -> typing.Type[pydantic.BaseModel]:
        """The model for the active operating mode."""
        return self.__models.get(self.active_mode)

    @property
    def model_schema(self) -> dict[str, typing.Any]:
        """The JSON schema representation of the model for the active operating mode."""
        return self.model_cls.model_json_schema()

    @property
    def flowgraph_cls(
        self,
    ) -> typing.Type[spectre_core.flowgraphs.Base]:
        """The flowgraph for the active operating mode."""
        return self.__flowgraphs.get(self.active_mode)

    @property
    def event_handler_cls(
        self,
    ) -> typing.Type[spectre_core.events.Base]:
        """The event handler for the active operating mode."""
        return self.__event_handlers.get(self.active_mode)

    @property
    def batch_cls(self) -> typing.Type[spectre_core.batches.Base]:
        """The batch for the active operating mode."""
        return self.__batches.get(self.active_mode)

    def model_validate(
        self, parameters: dict[str, typing.Any], skip: bool = False
    ) -> pydantic.BaseModel:
        """Create a model for the active operating mode using the input parameters.

        :param parameters: The parameters used to create the model.
        :param skip: If True, skip validating the parameters.
        :return: The validated model.
        """
        _LOGGER.info("Validating parameters...")
        return self.model_cls.model_validate(parameters, context={"skip": skip})

    def read_config(
        self,
        tag: str,
        skip_validation: bool = False,
        configs_dir_path: typing.Optional[str] = None,
    ) -> Config:
        """Read a config for this receiver.

        :param tag: The tag of the config.
        :param skip_validation: If True, skip validating the parameters.
        :param configs_dir_path: Optionally override the directory which stores the configs, defaults to None
        :return: A structure storing the file contents.
        :raises ValueError: If the config does not belong to this receiver.
        """
        configs_dir_path = (
            configs_dir_path or spectre_core.config.paths.get_configs_dir_path()
        )
        config = read_config(tag, configs_dir_path)
        if not config.receiver_name == self.name:
            raise ValueError(
                f"Config with tag {tag}, does not belong to this receiver. Expected '{self.name}', got {config.receiver_name}."
            )
        _ = self.model_validate(config.parameters, skip=skip_validation)
        return config

    def write_config(
        self,
        tag: str,
        parameters: dict[str, typing.Any],
        skip_validation: bool = False,
        configs_dir_path: typing.Optional[str] = None,
    ) -> None:
        """Write parameters to a config.

        :param tag: The tag of the config.
        :param parameters: The parameters to save.
        :param skip_validation: If True, skip validating the parameters.
        :param configs_dir_path: Optionally override the directory which stores the configs, defaults to None
        """
        write_config(
            tag,
            self.name,
            self.active_mode,
            self.model_validate(parameters, skip=skip_validation).model_dump(),
            configs_dir_path,
        )

    @spectre_core.logs.log_call
    def activate_flowgraph(
        self,
        tag: str,
        parameters: dict[str, typing.Any],
        skip_validation: bool = False,
        batches_dir_path: typing.Optional[str] = None,
    ) -> None:
        """Activate a flowgraph.

        :param config: The config used to configure the flowgraph.
        :param skip_validation: If True, skip validating the parameters.
        :param batches_dir_path: Optionally override the directory which stores the runtime data, defaults to None
        """
        _LOGGER.info("Starting the flowgraph...")
        self.flowgraph_cls(
            tag,
            self.model_validate(parameters, skip=skip_validation),
            batches_dir_path=batches_dir_path,
        ).activate()

    @spectre_core.logs.log_call
    def activate_post_processing(
        self,
        tag: str,
        parameters: dict[str, typing.Any],
        skip_validation: bool = False,
        batches_dir_path: typing.Optional[str] = None,
    ) -> None:
        """Activate post processing.

        :param config: The config used to configure post processing.
        :param skip_validation: If True, skip validating the parameters.
        :param batches_dir_path: Optionally override the directory which stores the runtime data, defaults to None
        """

        batches_dir_path = (
            batches_dir_path or spectre_core.config.paths.get_batches_dir_path()
        )
        observer = watchdog.observers.Observer()
        observer.schedule(
            self.event_handler_cls(
                tag,
                self.model_validate(parameters, skip=skip_validation),
                self.batch_cls,
            ),
            batches_dir_path,
            recursive=True,
            event_filter=[watchdog.events.FileCreatedEvent],
        )

        try:
            _LOGGER.info("Starting the post processing...")
            observer.start()
            observer.join()
        except KeyboardInterrupt:
            _LOGGER.warning(
                (
                    "Keyboard interrupt detected. Signalling "
                    "the post processing to stop"
                )
            )
            observer.stop()
            _LOGGER.warning(("Post processing has been successfully stopped"))

    def add_mode(
        self,
        mode: str,
        model: typing.Type[pydantic.BaseModel],
        flowgraph: typing.Type[spectre_core.flowgraphs.Base],
        event_handler: typing.Type[spectre_core.events.Base],
        batch: typing.Type[spectre_core.batches.Base],
    ) -> None:
        """Add an operating mode.

        :param mode: The name of the operating mode.
        :param model: The model defining configurable parameters for the operating mode.
        :param flowgraph: The flowgraph class for this operating mode, which will record
        the signal.
        :param event_handler: The event handler class for this operating mode, which will
        post process data recorded by the flowgraph.
        :param batch: The batch class used to read data produced at runtime by this operating mode.
        """
        self.__models.add(mode, model)
        self.__flowgraphs.add(mode, flowgraph)
        self.__event_handlers.add(mode, event_handler)
        self.__batches.add(mode, batch)



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import dataclasses
import typing

import numpy as np
import numpy.typing as npt
import astropy.io

import spectre_core.spectrograms
from ._base import Base, BatchFile


@dataclasses.dataclass(frozen=True)
class _Extension:
    FITS = "fits"


class _FitsFile(BatchFile[spectre_core.spectrograms.Spectrogram]):
    """Stores spectrogram data in the FITS format generated by the e-Callisto network."""

    def read(self) -> spectre_core.spectrograms.Spectrogram:
        """Parses a FITS file to generate a `Spectrogram`.

        Reverses the spectra along the frequency axis and converts units to linearised
        values if necessary. Infers the spectrum type from the `BUNIT` header.

        :raises NotImplementedError: If the `BUNIT` header value represents an unsupported spectrum type.
        :return: A `Spectrogram` instance containing the parsed FITS file data.
        """
        with astropy.io.fits.open(self.file_path, mode="readonly") as hdulist:
            primary_hdu = hdulist["PRIMARY"]

            date_obs = primary_hdu.header["DATE-OBS"]
            time_obs = primary_hdu.header["TIME-OBS"]
            spectrogram_start_datetime = datetime.datetime.strptime(
                f"{date_obs}T{time_obs}", "%Y/%m/%dT%H:%M:%S.%f"
            )

            bintable_hdu = hdulist[1]
            times = bintable_hdu.data["TIME"][0]
            frequencies = bintable_hdu.data["FREQUENCY"][0] * 1e6  # Convert to Hz

            bunit = primary_hdu.header["BUNIT"]

            # bunit is interpreted as a SpectrumUnit
            spectrum_unit = spectre_core.spectrograms.SpectrumUnit(bunit)
            if spectrum_unit == spectre_core.spectrograms.SpectrumUnit.DIGITS:
                dynamic_spectra_linearised = self._convert_units_to_linearised(
                    primary_hdu.data
                )

                return spectre_core.spectrograms.Spectrogram(
                    dynamic_spectra_linearised[
                        ::-1, :
                    ],  # reverse the spectra along the frequency axis
                    times,
                    frequencies[::-1],  # sort the frequencies in ascending order
                    spectrum_unit,
                    spectrogram_start_datetime,
                )
            else:
                raise NotImplementedError(
                    f"Spectrum type with BUNITS '{spectrum_unit}' is not currently supported."
                )

    def _convert_units_to_linearised(
        self, raw_digits: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """Converts spectrogram data from raw digit values to linearised units.

        Applies a transformation based on ADC specifications to convert raw values
        to dB and then to linearised units.

        :param dynamic_spectra: Raw dynamic spectra in digit values.
        :return: The dynamic spectra with linearised units.
        """
        # conversion as per ADC specs [see email from C. Monstein]
        dB = (raw_digits / 255) * (2500 / 25)
        return 10 ** (dB / 10)


class CallistoBatch(Base):
    def __init__(self, batches_dir_path: str, start_time: str, tag: str) -> None:
        """A batch of data generated by the e-Callisto network.

        Supports the following file extensions:
        - `.fits`

        :param start_time: The start time of the batch.
        :param tag: The batch name tag.
        """
        super().__init__(batches_dir_path, start_time, tag)

        self.add_file(_FitsFile, _Extension.FITS)

    @property
    def fits_file(self) -> _FitsFile:
        """The batch file corresponding to the `.fits` extension."""
        return typing.cast(_FitsFile, self.get_file(_Extension.FITS))

    @property
    def spectrogram_file(self) -> BatchFile[spectre_core.spectrograms.Spectrogram]:
        return self.fits_file



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import pydantic

import spectre_core.fields
import spectre_core.events
import spectre_core.flowgraphs

from ._validators import skip_validator, validate_window_size


class SignalGeneratorCosineWaveModel(
    spectre_core.flowgraphs.SignalGeneratorCosineWaveModel,
    spectre_core.events.FixedCenterFrequencyModel,
):
    window_type: spectre_core.fields.Field.window_type = "boxcar"

    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_window_size(self.window_size)

        if not self.window_type == "boxcar":
            raise ValueError(
                f"The window type must be boxcar. Got '{self.window_type}'"
            )
        if self.sample_rate % self.frequency != 0:
            raise ValueError(
                "The sampling rate must be some integer multiple of frequency"
            )

        a = self.sample_rate / self.frequency
        if a < 2:
            raise ValueError(
                (
                    f"The ratio of sampling rate over frequency must be greater than two. "
                    f"Got {a}"
                )
            )
        p = self.window_size / a
        if self.window_size % a != 0:
            raise ValueError(
                f"The number of sampled cycles must be a positive natural number. Computed p={p}."
            )

        return self


class SignalGeneratorConstantStaircaseModel(
    spectre_core.flowgraphs.SignalGeneratorConstantStaircaseModel,
    spectre_core.events.FixedCenterFrequencyModel,
):
    window_type: spectre_core.fields.Field.window_type = "boxcar"

    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_window_size(self.window_size)
        if not self.window_type == "boxcar":
            raise ValueError(
                f"The window type must be boxcar. Got '{self.window_type}'"
            )
        if self.frequency_step != self.sample_rate:
            raise ValueError(f"The frequency step must be equal to the sampling rate")

        if self.min_samples_per_step > self.max_samples_per_step:
            raise ValueError(
                (
                    f"Minimum samples per step cannot be greater than the maximum samples per step. "
                    f"Got {self.min_samples_per_step}, which is greater than {self.max_samples_per_step}"
                )
            )
        return self



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from gnuradio import spectre
from gnuradio import soapy

import spectre_core.fields

from ._base import Base, BaseModel


class RTLSDRFixedCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 1536000
    center_frequency: spectre_core.fields.Field.center_frequency = 95.8e6
    rf_gain: spectre_core.fields.Field.rf_gain = 30
    batch_size: spectre_core.fields.Field.batch_size = 3


class RTLSDRFixedCenterFrequency(Base[RTLSDRFixedCenterFrequencyModel]):
    def configure(self, tag: str, model: RTLSDRFixedCenterFrequencyModel) -> None:
        stream_args = ""
        tune_args = [""]
        settings = [""]
        self.soapy_rtlsdr_source = soapy.source(
            "driver=rtlsdr", "fc32", 1, "", stream_args, tune_args, settings
        )
        self.soapy_rtlsdr_source.set_sample_rate(0, model.sample_rate)
        self.soapy_rtlsdr_source.set_gain_mode(0, False)
        self.soapy_rtlsdr_source.set_frequency(0, model.center_frequency)
        self.soapy_rtlsdr_source.set_frequency_correction(0, 0)
        self.soapy_rtlsdr_source.set_gain(0, model.rf_gain)
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.batch_size,
            model.sample_rate,
            False,
            "rx_freq",
            0,
        )
        self.connect((self.soapy_rtlsdr_source, 0), (self.spectre_batched_file_sink, 0))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing

import numpy as np
import numpy.typing as npt


def average_array(
    array: npt.NDArray[np.float32], average_over: int, axis: int = 0
) -> npt.NDArray[np.float32]:
    """
    Averages elements of an array in blocks along a specified axis.

    :param array: Input array to be averaged.
    :param average_over: Number of elements in each averaging block.
    :param axis: Axis along which to perform the averaging, defaults to 0.
    :raises TypeError: If `average_over` is not an integer.
    :raises ValueError: If `average_over` is not in the range [1, size of the axis].
    :raises ValueError: If `axis` is out of bounds for the array.
    :return: Array of averaged values along the specified axis.
    """

    # Get the size of the specified axis which we will average over
    axis_size = array.shape[axis]
    # Check if average_over is within the valid range
    if not 1 <= average_over <= axis_size:
        raise ValueError(
            f"average_over must be between 1 and the length of the axis ({axis_size})"
        )

    max_axis_index = len(np.shape(array)) - 1
    if axis > max_axis_index:  # zero indexing on specifying axis, so minus one
        raise ValueError(
            f"Requested axis is out of range of array dimensions. Axis: {axis}, max axis index: {max_axis_index}"
        )

    # find the number of elements in the requested axis
    num_elements = array.shape[axis]

    # find the number of "full blocks" to average over
    num_full_blocks = num_elements // average_over
    # if num_elements is not exactly divisible by average_over, we will have some elements left over
    # these remaining elements will be padded with nans to become another full block
    remainder = num_elements % average_over

    # if there exists a remainder, pad the last block
    if remainder != 0:
        # initialise an array to hold the padding shape
        padding_shape = [(0, 0)] * array.ndim
        # pad after the last column in the requested axis
        padding_shape[axis] = (0, average_over - remainder)
        # pad with nan values (so to not contribute towards the mean computation)
        array = np.pad(array, padding_shape, mode="constant", constant_values=np.nan)

    # initalise a list to hold the new shape
    new_shape = list(array.shape)
    # update the shape on the requested access (to the number of blocks we will average over)
    new_shape[axis] = num_full_blocks + (1 if remainder else 0)
    # insert a new dimension, with the size of each block
    new_shape.insert(axis + 1, average_over)
    # and reshape the array to sort the array into the relevant blocks.
    reshaped_array = array.reshape(new_shape)
    # average over the newly created axis, essentially averaging over the blocks.
    averaged_array = np.nanmean(reshaped_array, axis=axis + 1)
    # return the averaged array
    return averaged_array


T = typing.TypeVar("T", np.float32, np.datetime64)


def find_closest_index(
    target_value: T, array: npt.NDArray[T], enforce_strict_bounds: bool = False
) -> int:
    """
    Finds the index of the closest value to a target in a given array, with optional bounds enforcement.

    :param target_value: The value to find the closest match for.
    :param array: The array to search within.
    :param enforce_strict_bounds: If True, raises an error if the target value is outside the array bounds. Defaults to False.
    :return: The index of the closest value in the array.
    :raises ValueError: If `enforce_strict_bounds` is True and `target_value` is outside the array bounds.
    """
    # Check bounds if strict enforcement is required
    if enforce_strict_bounds:
        max_value, min_value = np.nanmax(array), np.nanmin(array)
        if target_value > max_value:
            raise ValueError(
                f"Target value {target_value} exceeds max array value {max_value}"
            )
        if target_value < min_value:
            raise ValueError(
                f"Target value {target_value} is less than min array value {min_value}"
            )

    # Find the index of the closest value
    return int(np.argmin(np.abs(array - target_value)))


def normalise_peak_intensity(array: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
    """
    Normalises an array by its peak intensity.

    :param array: Input array to normalise.
    :return: Array normalised such that its maximum value is 1. NaN values are ignored.
    """
    return array / np.nanmax(array)


def compute_resolution(array: npt.NDArray[np.float32]) -> float:
    """
    Computes the median resolution of a one-dimensional array.

    :param array: Input one-dimensional array of values.
    :return: The median of differences between consecutive elements in the array.
    :raises ValueError: If the input array is not one-dimensional or contains fewer than two elements.
    """
    # Check that the array is one-dimensional
    if array.ndim != 1:
        raise ValueError("Input array must be one-dimensional")

    if len(array) < 2:
        raise ValueError("Input array must contain at least two elements")

    # Calculate differences between consecutive elements.
    resolutions = np.diff(array)
    return float(np.nanmedian(resolutions))


def compute_range(array: npt.NDArray[np.float32]) -> float:
    """
    Computes the range of a one-dimensional array as the difference between its last and first elements.

    :param array: Input one-dimensional array of values.
    :return: The range of the array (last element minus first element).
    :raises ValueError: If the input array is not one-dimensional or contains fewer than two elements.
    """
    # Check that the array is one-dimensional
    if array.ndim != 1:
        raise ValueError("Input array must be one-dimensional")

    if len(array) < 2:
        raise ValueError("Input array must contain at least two elements")
    return float(array[-1] - array[0])


def subtract_background(
    array: npt.NDArray[np.float32], start_index: int, end_index: int
) -> npt.NDArray[np.float32]:
    """
    Subtracts the mean of a specified background range from all elements in an array.

    :param array: Input array from which the background mean will be subtracted.
    :param start_index: Start index of the background range (inclusive).
    :param end_index: End index of the background range (inclusive).
    :return: Array with the background mean subtracted.
    """
    array -= np.nanmean(array[start_index : end_index + 1])
    return array


def time_elapsed(datetimes: npt.NDArray[np.datetime64]) -> npt.NDArray[np.float32]:
    """Convert an array of datetimes to seconds elapsed."""
    return (datetimes - datetimes[0]).astype("timedelta64[us]") / np.timedelta64(1, "s")



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing
import datetime

import matplotlib.colors
import matplotlib.cm
import numpy as np
import numpy.typing as npt

import spectre_core.spectrograms
from ._base import BasePanel, BaseTimeSeriesPanel, XAxisType
from ._panel_names import PanelName


T = typing.TypeVar("T")


def _bind_to_colors(
    values: list[T], cmap: str = "winter"
) -> typing.Iterator[tuple[T, npt.NDArray[np.float32]]]:
    """
    Assign RGBA colors to a list of values using a colormap.

    Each value is mapped linearly to a subset of the unit interval and then converted
    to an RGBA color using the specified colormap.

    :param values: List of values to map to colors.
    :param cmap: Name of the Matplotlib colormap to use. Defaults to "winter".
    :return: An iterator of tuples, each containing a value and its corresponding RGBA color.
    """
    colormap = matplotlib.cm.get_cmap(cmap)
    rgbas = colormap(np.linspace(0.1, 0.9, len(values)))
    return zip(values, rgbas)


class FrequencyCutsPanel(BasePanel):
    """
    Panel for visualising spectrogram data as frequency cuts.

    This panel plots spectrums corresponding to specific time instances
    in the spectrogram. Each cut is drawn as a line plot, optionally normalized
    or converted to decibels above the background.
    """

    def __init__(
        self,
        spectrogram: spectre_core.spectrograms.Spectrogram,
        *times: float | str,
        dBb: bool = False,
        peak_normalise: bool = False,
    ) -> None:
        """Initialise an instance of `FrequencyCutsPanel`.

        :param spectrogram: The spectrogram to be visualised.
        :param *times: Times at which to take frequency cuts. Can be floats (relative time) or
        strings (formatted datetimes).
        :param dBb: If True, plots the spectrums in decibels above the background. Defaults to False.
        :param peak_normalise: If True, normalizes each spectrum such that its peak value is 1.
        Ignored if `dBb` is True. Defaults to False.
        """
        super().__init__(PanelName.FREQUENCY_CUTS, spectrogram)

        if len(times) == 0:
            raise ValueError(
                f"You must specify the time of at least one cut in `*times`"
            )
        self._times = times

        self._dBb = dBb
        self._peak_normalise = peak_normalise
        self._frequency_cuts: dict[
            float | datetime.datetime, spectre_core.spectrograms.FrequencyCut
        ] = {}

    @property
    def xaxis_type(self) -> typing.Literal[XAxisType.FREQUENCY]:
        return XAxisType.FREQUENCY

    @property
    def frequencies(self) -> npt.NDArray[np.float32]:
        """The physical frequencies assigned to each spectral component."""
        return self._spectrogram.frequencies

    def annotate_xaxis(self) -> None:
        """Annotate the x-axis assuming frequency in units of Hz."""
        self._get_ax().set_xlabel("Frequency [Hz]")

    def get_frequency_cuts(
        self,
    ) -> dict[float | datetime.datetime, spectre_core.spectrograms.FrequencyCut]:
        """
        Get the frequency cuts for the specified times.

        Computes and caches the spectrum for each requested time. The results are
        stored as a mapping from time to the corresponding `FrequencyCut`.

        :return: A dictionary mapping each time to its corresponding frequency cut.
        """
        if not self._frequency_cuts:
            for time in self._times:
                frequency_cut = self._spectrogram.get_frequency_cut(
                    time, dBb=self._dBb, peak_normalise=self._peak_normalise
                )
                self._frequency_cuts[frequency_cut.time] = frequency_cut
        return self._frequency_cuts

    def get_cut_times(self) -> list[float | datetime.datetime]:
        """
        Get the exact times of the frequency cuts.

        Returns the closest matches to the times specified in the constructor.

        :return: A list of times corresponding to the stored frequency cuts.
        """
        frequency_cuts = self.get_frequency_cuts()
        return list(frequency_cuts.keys())

    def draw(self) -> None:
        """Draw the frequency cuts onto the panel."""
        frequency_cuts = self.get_frequency_cuts()
        for time, color in self.bind_to_colors():
            frequency_cut = frequency_cuts[time]
            self._get_ax().step(
                self.frequencies,  # convert to MHz
                frequency_cut.cut,
                where="mid",
                color=color,
            )

    def annotate_yaxis(self) -> None:
        """Annotate the y-axis of the panel based on the current state.

        The y-axis label reflects whether the data is in decibels above the background (`dBb`),
        normalized to peak values, or in the original units of the spectrogram.
        """
        ax = self._get_ax()
        if self._dBb:
            ax.set_ylabel("dBb")
        elif self._peak_normalise:
            # no y-axis label if we are peak normalising
            return
        else:
            ax.set_ylabel(f"{self._spectrogram.spectrum_unit.value.capitalize()}")

    def bind_to_colors(
        self,
    ) -> typing.Iterator[tuple[float | datetime.datetime, npt.NDArray[np.float32]]]:
        """
        Bind each frequency cut time to an RGBA color.

        Colors are assigned using the colormap specified in the panel format.

        :return: An iterator of tuples, each containing a cut time and its corresponding RGBA color.
        """
        return _bind_to_colors(
            self.get_cut_times(), cmap=self.get_panel_format().line_cmap
        )


class TimeCutsPanel(BaseTimeSeriesPanel):
    """
    Panel for visualising spectrogram data as time series of spectral components.

    This panel plots the time evolution of spectral components at specific
    frequencies in the spectrogram. Each time series is drawn as a line plot,
    optionally normalized, background-subtracted, or converted to decibels above the background.
    """

    def __init__(
        self,
        spectrogram: spectre_core.spectrograms.Spectrogram,
        *frequencies: float,
        dBb: bool = False,
        peak_normalise: bool = False,
        background_subtract: bool = False,
    ) -> None:
        """Initialise an instance of `TimeCutsPanel`.

        :param spectrogram: The spectrogram to be visualised.
        :param *frequencies: Frequencies at which to extract time series.
        :param dBb: If True, returns the cuts in decibels above the background. Defaults to False.
        :param peak_normalise: If True, normalizes each time series so its peak value is 1.
        Ignored if `dBb` is True. Defaults to False.
        :param background_subtract: If True, subtracts the background from each time series.
        Ignored if `dBb` is True. Defaults to False.
        """
        super().__init__(PanelName.TIME_CUTS, spectrogram)

        if len(frequencies) == 0:
            raise ValueError(
                f"You must specify the frequency of at least one cut in `*frequencies`."
            )
        self._frequencies = frequencies

        self._dBb = dBb
        self._peak_normalise = peak_normalise
        self._background_subtract = background_subtract
        self._time_cuts: dict[float, spectre_core.spectrograms.TimeCut] = {}

    def get_time_cuts(self) -> dict[float, spectre_core.spectrograms.TimeCut]:
        """
        Get the time cuts for the specified frequencies.

        Computes and caches the time series for each requested frequency. The results
        are stored as a mapping from frequency to `TimeCut`.

        :return: A dictionary mapping each frequency to its corresponding time cut.
        """
        if not self._time_cuts:
            for frequency in self._frequencies:
                time_cut = self._spectrogram.get_time_cut(
                    frequency,
                    dBb=self._dBb,
                    peak_normalise=self._peak_normalise,
                    correct_background=self._background_subtract,
                    return_time_type=self.get_time_type(),
                )
                self._time_cuts[time_cut.frequency] = time_cut
        return self._time_cuts

    def get_frequencies(self) -> list[float]:
        """
        Get the exact frequencies for the spectral components being plotted.

        :return: A list of frequencies corresponding to the stored time cuts.
        """
        time_cuts = self.get_time_cuts()
        return list(time_cuts.keys())

    def draw(self) -> None:
        """Draw the time series for each spectral component onto the panel."""
        time_cuts = self.get_time_cuts()
        for frequency, color in self.bind_to_colors():
            time_cut = time_cuts[frequency]
            self._get_ax().step(self.times, time_cut.cut, where="mid", color=color)

    def annotate_yaxis(self) -> None:
        """
        Annotate the y-axis of the panel based on the current state.

        The y-axis label reflects whether the data is in decibels above the background (`dBb`),
        normalized to peak values, or in the original units of the spectrogram.
        """
        ax = self._get_ax()
        if self._dBb:
            ax.set_ylabel("dBb")
        elif self._peak_normalise:
            return  # no y-axis label if we are peak normalising.
        else:
            ax.set_ylabel(f"{self._spectrogram.spectrum_unit.value.capitalize()}")

    def bind_to_colors(self) -> typing.Iterator[tuple[float, npt.NDArray[np.float32]]]:
        """
        Bind each spectral component's frequency to an RGBA color.

        Colors are assigned using the colormap specified in the panel format.

        :return: An iterator of tuples, each containing a frequency and its corresponding RGBA color.
        """
        return _bind_to_colors(
            self.get_frequencies(), cmap=self.get_panel_format().line_cmap
        )


class IntegralOverFrequencyPanel(BaseTimeSeriesPanel):
    """Panel for visualising the spectrogram integrated over frequency.

    This panel plots the spectrogram numerically integrated over frequency as a time
    series. The result can be normalized to its peak value or adjusted by subtracting
    the background.
    """

    def __init__(
        self,
        spectrogram: spectre_core.spectrograms.Spectrogram,
        peak_normalise: bool = False,
        background_subtract: bool = False,
    ):
        """Initialise an instance of `IntegralOverFrequencyPanel`.

        :param spectrogram: The spectrogram to be visualised.
        :param peak_normalise: If True, normalizes the integral so its peak value is 1. Defaults to False.
        :param background_subtract: If True, subtracts the background after computing the integral. Defaults to False.
        """
        super().__init__(PanelName.INTEGRAL_OVER_FREQUENCY, spectrogram)
        self._peak_normalise = peak_normalise
        self._background_subtract = background_subtract

    def draw(self):
        """Integrate the spectrogram over frequency and plot the result."""
        I = self._spectrogram.integrate_over_frequency(
            correct_background=self._background_subtract,
            peak_normalise=self._peak_normalise,
        )
        self._get_ax().step(
            self.times, I, where="mid", color=self.get_panel_format().line_color
        )

    def annotate_yaxis(self):
        """This panel does not annotate the y-axis."""


class SpectrogramPanel(BaseTimeSeriesPanel):
    """
    Panel for visualising the full spectrogram.

    This panel plots the spectrogram as a colormap, with optional log normalization or
    in units of decibels above the background.
    """

    def __init__(
        self,
        spectrogram: spectre_core.spectrograms.Spectrogram,
        log_norm: bool = False,
        dBb: bool = False,
        vmin: typing.Optional[float] = None,
        vmax: typing.Optional[float] = None,
    ) -> None:
        """Initialise an instance of `SpectrogramPanel`.

        :param spectrogram: The spectrogram to be visualised.
        :param log_norm: If True, normalizes the values to the 0-1 range on a logarithmic scale. Defaults to False.
        :param dBb: If True, plots the spectrogram in decibels above the background. Defaults to False.
        :param vmin: Minimum value for the colormap. Only applies if `dBb` is True. Defaults to None.
        :param vmax: Maximum value for the colormap. Only applies if `dBb` is True. Defaults to None.
        """
        super().__init__(PanelName.SPECTROGRAM, spectrogram)
        self._log_norm = log_norm
        self._dBb = dBb
        self._vmin = vmin
        self._vmax = vmax

    def _draw_dBb(self) -> None:
        """Plot the spectrogram in decibels above the background (dBb).

        This method handles plotting the spectrogram with dBb scaling, applying
        colormap bounds (`vmin` and `vmax`) and adding a colorbar to the panel.
        """
        dynamic_spectra = self._spectrogram.compute_dynamic_spectra_dBb()

        # use defaults if neither vmin or vmax is specified
        vmin = self._vmin or -1
        vmax = self._vmax or 2

        ax = self._get_ax()
        # Plot the spectrogram
        pcm = ax.pcolormesh(
            self.times,
            self._spectrogram.frequencies,
            dynamic_spectra,
            vmin=vmin,
            vmax=vmax,
            cmap=self.get_panel_format().spectrogram_cmap,
        )

        # Add colorbar
        cbar_ticks = np.linspace(vmin, vmax, 6)
        cbar = self._get_fig().colorbar(pcm, ax=ax, ticks=cbar_ticks)
        cbar.set_label("dBb")

    def _draw_normal(self) -> None:
        """Plot the spectrogram with optional logarithmic normalization.

        This method handles plotting the spectrogram without dBb scaling, using
        linear or log normalization based on the `log_norm` attribute.
        """
        dynamic_spectra = self._spectrogram.dynamic_spectra

        if self._log_norm:
            norm = matplotlib.colors.LogNorm(
                vmin=np.nanmin(dynamic_spectra[dynamic_spectra > 0]),
                vmax=np.nanmax(dynamic_spectra),
            )
        else:
            norm = None

        # Plot the spectrogram
        self._get_ax().pcolormesh(
            self.times,
            self._spectrogram.frequencies,
            dynamic_spectra,
            cmap=self.get_panel_format().spectrogram_cmap,
            norm=norm,
        )

    def draw(self) -> None:
        """Draw the spectrogram onto the panel."""
        if self._dBb:
            self._draw_dBb()
        else:
            self._draw_normal()

    def annotate_yaxis(self) -> None:
        """Annotate the yaxis, assuming units of Hz."""
        self._get_ax().set_ylabel("Frequency [Hz]")
        return

    def overlay_time_cuts(self, cuts_panel: TimeCutsPanel) -> None:
        """
        Overlay horizontal lines on the spectrogram to indicate time cuts.

        The lines correspond to the frequencies of the cuts on a `TimeCutsPanel`.
        Colors are matched to the lines on the `TimeCutsPanel`.

        :param cuts_panel: The `TimeCutsPanel` containing the cut frequencies to overlay.
        """
        for frequency, color in cuts_panel.bind_to_colors():
            self._get_ax().axhline(
                frequency, color=color, linewidth=self.get_panel_format().line_width
            )

    def overlay_frequency_cuts(self, cuts_panel: FrequencyCutsPanel) -> None:
        """
        Overlay vertical lines on the spectrogram to indicate frequency cuts.

        The lines correspond to the times of the cuts on a `FrequencyCutsPanel`.
        Colors are matched to the lines on the `FrequencyCutsPanel`.

        :param cuts_panel: The `FrequencyCutsPanel` containing the cut times to overlay.
        """
        for time, color in cuts_panel.bind_to_colors():
            self._get_ax().axvline(
                time, color=color, linewidth=self.get_panel_format().line_width
            )



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""A vendor-neutral interface for recording signals and spectrograms from SDRs."""

from ._register import register_receiver, get_registered_receivers
from ._factory import get_receiver, get_batch_cls
from ._config import (
    Config,
    read_config,
    write_config,
    get_config_file_path,
    parse_config_file_name,
)
from ._base import Base
from ._names import ReceiverName
from ._custom import Custom
from ._signal_generator import SignalGenerator
from ._rsp1a import RSP1A
from ._rspduo import RSPduo
from ._rspdx import RSPdx
from ._usrp import USRP
from ._b200mini import B200mini
from ._hackrf import HackRF
from ._hackrfone import HackRFOne
from ._rtlsdr import RTLSDR
from ._record import record_signal, record_spectrograms

__all__ = [
    "register_receiver",
    "get_registered_receivers",
    "get_receiver",
    "get_batch_cls",
    "get_config_file_path",
    "parse_config_file_name",
    "ReceiverName",
    "Config",
    "read_config",
    "write_config",
    "Base",
    "Custom",
    "SignalGenerator",
    "RSP1A",
    "RSPduo",
    "RSPdx",
    "USRP",
    "B200mini",
    "HackRF",
    "HackRFOne",
    "RTLSDR",
    "ReceiverName",
    "record_signal",
    "record_spectrograms",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
import datetime
import typing
import math

from ._array_operations import find_closest_index, average_array, time_elapsed
from ._spectrogram import Spectrogram


def frequency_chop(
    spectrogram: Spectrogram, start_frequency: float, end_frequency: float
) -> Spectrogram:
    """
    Extracts a portion of the spectrogram within the specified frequency range.

    :param spectrogram: The input spectrogram to process.
    :param start_frequency: The starting frequency of the desired range (Hz).
    :param end_frequency: The ending frequency of the desired range (Hz).
    :raises ValueError: If the specified frequency range is entirely outside the spectrogram's frequency range.
    :raises ValueError: If the start and end indices for the frequency range are identical.
    :return: A new spectrogram containing only the specified frequency range.
    """
    is_entirely_below_frequency_range = (
        start_frequency < spectrogram.frequencies[0]
        and end_frequency < spectrogram.frequencies[0]
    )
    is_entirely_above_frequency_range = (
        start_frequency > spectrogram.frequencies[-1]
        and end_frequency > spectrogram.frequencies[-1]
    )
    if is_entirely_below_frequency_range or is_entirely_above_frequency_range:
        raise ValueError(
            f"The requested frequency interval is entirely out of range of the input spectrogram."
        )

    # find the index of the nearest matching frequency bins in the spectrogram
    start_index = find_closest_index(
        np.float32(start_frequency), spectrogram.frequencies
    )
    end_index = find_closest_index(np.float32(end_frequency), spectrogram.frequencies)

    # enforce distinct start and end indices
    if start_index == end_index:
        raise ValueError(
            f"Start and end indices are equal! Got start_index: {start_index} and end_index: {end_index}"
        )

    # if start index is more than end index, swap the ordering so to enforce start_index <= end_index
    if start_index > end_index:
        start_index, end_index = end_index, start_index

    # chop the spectrogram accordingly
    transformed_dynamic_spectra = spectrogram.dynamic_spectra[
        start_index : end_index + 1, :
    ]
    transformed_frequencies = spectrogram.frequencies[start_index : end_index + 1]

    return Spectrogram(
        transformed_dynamic_spectra,
        spectrogram.times,
        transformed_frequencies,
        spectrogram.spectrum_unit,
        spectrogram.start_datetime,
    )


def time_chop(
    spectrogram: Spectrogram,
    start_datetime: datetime.datetime,
    end_datetime: datetime.datetime,
) -> Spectrogram:
    """
    Extracts a portion of the spectrogram within the specified time range.

    :param spectrogram: The input spectrogram to process.
    :param start_datetime: The starting time of the desired range.
    :param end_datetime: The ending time of the desired range.
    :raises ValueError: If the specified time range is entirely outside the spectrogram's time range.
    :raises ValueError: If the start and end indices for the time range are identical.
    :return: A new spectrogram containing only the specified time range.
    """
    start_datetime64 = np.datetime64(start_datetime)
    end_datetime64 = np.datetime64(end_datetime)

    is_entirely_below_time_range = (
        start_datetime64 < spectrogram.datetimes[0]
        and end_datetime64 < spectrogram.datetimes[0]
    )
    is_entirely_above_time_range = (
        start_datetime64 > spectrogram.datetimes[-1]
        and end_datetime64 > spectrogram.datetimes[-1]
    )
    if is_entirely_below_time_range or is_entirely_above_time_range:
        raise ValueError(
            f"Requested time interval is entirely out of range of the input spectrogram."
        )

    # find the index of the nearest matching spectrums in the spectrogram.
    start_index = find_closest_index(start_datetime64, spectrogram.datetimes)
    end_index = find_closest_index(end_datetime64, spectrogram.datetimes)

    # enforce distinct start and end indices
    if start_index == end_index:
        raise ValueError(
            f"Start and end indices are equal! Got start_index: {start_index} and end_index: {end_index}"
        )

    # if start index is more than end index, swap the ordering so to enforce start_index <= end_index
    if start_index > end_index:
        start_index, end_index = end_index, start_index

    # chop the spectrogram accordingly
    transformed_dynamic_spectra = spectrogram.dynamic_spectra[
        :, start_index : end_index + 1
    ]
    transformed_start_datetime = spectrogram.datetimes[start_index]

    # chop the times array and translate such that the first spectrum to t=0 [s]
    transformed_times = spectrogram.times[start_index : end_index + 1]
    transformed_times -= transformed_times[0]

    return Spectrogram(
        transformed_dynamic_spectra,
        transformed_times,
        spectrogram.frequencies,
        spectrogram.spectrum_unit,
        transformed_start_datetime,
    )


def _validate_and_compute_average_over(
    original_resolution: float, resolution: typing.Optional[float], average_over: int
) -> int:
    """
    Validates the input parameters and computes `average_over` if `resolution` is specified.

    :param resolution: The desired resolution for averaging. Mutually exclusive with `average_over`.
    :param average_over: The number of consecutive spectrums to average over. Mutually exclusive with `resolution`.
    :param original_resolution: The original resolution (e.g., time or frequency).
    :raises ValueError: If neither or both `resolution` and `average_over` are specified.
    :return: The computed or validated `average_over` value.
    """
    if (resolution is None) and (average_over == 1):
        return average_over

    if not (resolution is not None) ^ (average_over != 1):
        raise ValueError(
            "Exactly one of 'resolution' or 'average_over' must be specified."
        )

    if resolution is not None:
        return max(1, math.floor(resolution / original_resolution))

    else:
        return average_over


def time_average(
    spectrogram: Spectrogram,
    resolution: typing.Optional[float] = None,
    average_over: int = 1,
) -> Spectrogram:
    """
    Performs time averaging on the spectrogram data.

    :param spectrogram: The input spectrogram to process.
    :param resolution: The desired time resolution for averaging (seconds). Mutually exclusive with `average_over`.
    :param average_over: The number of consecutive time points to average. Mutually exclusive with `resolution`.
    :raises NotImplementedError: If the spectrogram lacks a defined start datetime.
    :raises ValueError: If neither or both `resolution` and `average_over` are specified.
    :return: A new spectrogram with time-averaged data.
    """
    if not spectrogram.start_datetime_is_set:
        raise NotImplementedError(
            "Time averaging is not supported for spectrograms without an assigned start datetime."
        )

    average_over = _validate_and_compute_average_over(
        spectrogram.time_resolution, resolution, average_over
    )

    # Perform averaging
    transformed_dynamic_spectra = average_array(
        spectrogram.dynamic_spectra, average_over, axis=1
    )
    transformed_times = average_array(spectrogram.times, average_over)

    # Update start datetime and adjust times to start at t=0
    transformed_start_datetime = spectrogram.datetimes[0] + (
        transformed_times[0] * 1e6
    ).astype("timedelta64[us]")
    transformed_times -= transformed_times[0]

    return Spectrogram(
        transformed_dynamic_spectra,
        transformed_times,
        spectrogram.frequencies,
        spectrogram.spectrum_unit,
        transformed_start_datetime,
    )


def frequency_average(
    spectrogram: Spectrogram,
    resolution: typing.Optional[float] = None,
    average_over: int = 1,
) -> Spectrogram:
    """
    Performs frequency averaging on the spectrogram data.

    :param spectrogram: The input spectrogram to process.
    :param resolution: The desired frequency resolution for averaging (Hz). Mutually exclusive with `average_over`.
    :param average_over: The number of consecutive frequency bins to average. Mutually exclusive with `resolution`.
    :raises ValueError: If neither or both `resolution` and `average_over` are specified.
    :return: A new spectrogram with frequency-averaged data.
    """
    average_over = _validate_and_compute_average_over(
        spectrogram.frequency_resolution, resolution, average_over
    )

    # Perform averaging
    transformed_dynamic_spectra = average_array(
        spectrogram.dynamic_spectra, average_over, axis=0
    )
    transformed_frequencies = average_array(spectrogram.frequencies, average_over)

    return Spectrogram(
        transformed_dynamic_spectra,
        spectrogram.times,
        transformed_frequencies,
        spectrogram.spectrum_unit,
        spectrogram.start_datetime,
    )


def join_spectrograms(spectrograms: list[Spectrogram]) -> Spectrogram:
    """
    Joins multiple spectrograms into a single spectrogram along the time axis.

    :param spectrograms: A list of spectrograms to combine.
    :raises ValueError: If the input list is empty.
    :raises ValueError: If spectrograms have mismatched frequency ranges.
    :raises ValueError: If spectrograms have different tags.
    :raises ValueError: If spectrograms have differing spectrum units.
    :raises ValueError: If any spectrogram lacks a defined start datetime.
    :return: A new spectrogram combining all input spectrograms along the time axis.
    """
    # check that the length of the list is non-zero
    num_spectrograms = len(spectrograms)
    if num_spectrograms == 0:
        raise ValueError(f"Input list of spectrograms is empty!")

    # extract the first element of the list, and use this as a reference for comparison
    # input validations.
    reference_spectrogram = spectrograms[0]

    # perform checks on each spectrogram in teh list
    for spectrogram in spectrograms:
        if not np.all(
            np.equal(spectrogram.frequencies, reference_spectrogram.frequencies)
        ):
            raise ValueError(f"All spectrograms must have identical frequency ranges")
        if spectrogram.spectrum_unit != reference_spectrogram.spectrum_unit:
            raise ValueError(
                f"All units must be equal for each spectrogram in the input list!"
            )
        if not spectrogram.start_datetime_is_set:
            raise ValueError(f"All spectrograms must have their start datetime set.")

    # Concatenate all dynamic spectra directly along the time axis
    transformed_dynamic_spectra = np.hstack(
        [spectrogram.dynamic_spectra for spectrogram in spectrograms]
    )

    transformed_times = time_elapsed(
        np.concatenate([s.datetimes for s in spectrograms])
    )

    return Spectrogram(
        transformed_dynamic_spectra,
        transformed_times,
        reference_spectrogram.frequencies,
        reference_spectrogram.spectrum_unit,
        reference_spectrogram.start_datetime,
    )



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing
import warnings
import datetime
import dataclasses
import enum
import os

import numpy as np
import numpy.typing as npt
import astropy.io.fits

import spectre_core.config
from ._array_operations import (
    find_closest_index,
    normalise_peak_intensity,
    compute_resolution,
    compute_range,
    subtract_background,
)


class SpectrumUnit(enum.Enum):
    """A defined unit for dynamic spectra values.

    :ivar AMPLITUDE: Formal definition TBC.
    :ivar POWER: Formal definition TBC.
    :ivar DIGITS: Formal definition TBC.
    """

    AMPLITUDE = "amplitude"
    POWER = "power"
    DIGITS = "digits"


@dataclasses.dataclass
class FrequencyCut:
    """A cut of a dynamic spectra, at a particular instant of time. Equivalently, some spectrum in
    the spectrogram.

    :ivar time: The time of the frequency cut, either in relative time (if time is a float)
    or as a datetime.
    :ivar frequencies: The physical frequencies assigned to each spectral component, in Hz.
    :ivar cut: The spectrum values.
    :ivar spectrum_unit: The unit of each spectrum value.
    """

    time: float | datetime.datetime
    frequencies: npt.NDArray[np.float32]
    cut: npt.NDArray[np.float32]
    spectrum_unit: SpectrumUnit


@dataclasses.dataclass
class TimeCut:
    """A cut of a dynamic spectra, at some fixed frequency. Equivalently, a time series of
    some spectral component in the spectrogram.

    :ivar frequency: The physical frequency assigned to the spectral component, in Hz.
    :ivar times: The time for each time series value, either as a relative time (if
    the elements are floats) or as a datetimes.
    :ivar cut: The time series values of the spectral component.
    :ivar spectrum_unit: The unit of each time series value.
    """

    frequency: float
    times: npt.NDArray[np.float32 | np.datetime64]
    cut: npt.NDArray[np.float32]
    spectrum_unit: SpectrumUnit


class TimeType(enum.Enum):
    """The type of time we can assign to each spectrum in the dynamic spectra.

    :ivar RELATIVE: The elapsed time from the first spectrum, in seconds.
    :ivar DATETIMES: The datetime associated with each spectrum.
    """

    RELATIVE = "relative"
    DATETIMES = "datetimes"


class Spectrogram:
    """Standardised wrapper for spectrogram data, providing a consistent
    interface for storing, accessing, and manipulating spectrogram data,
    along with associated metadata.
    """

    def __init__(
        self,
        dynamic_spectra: npt.NDArray[np.float32],
        times: npt.NDArray[np.float32],
        frequencies: npt.NDArray[np.float32],
        spectrum_unit: SpectrumUnit,
        start_datetime: typing.Optional[datetime.datetime | np.datetime64] = None,
    ) -> None:
        """Initialise a Spectrogram instance.

        :param dynamic_spectra: A 2D array of spectrogram data.
        :param times: A 1D array representing the elapsed time of each spectrum, in seconds, relative to the first
        in the spectrogram.
        :param frequencies: A 1D array representing the physical frequencies assigned to each spectral component, in Hz.
        :param tag: A string identifier for the spectrogram.
        :param spectrum_unit: The unit of the dynamic_spectra values.
        :param start_datetime: The datetime corresponding to the first spectrum, defaults to None.
        :raises ValueError: If times does not start at 0 or array shapes are inconsistent.
        """

        self._dynamic_spectra = dynamic_spectra

        if times[0] != 0:
            raise ValueError(f"The first spectrum must correspond to t=0")
        self._times = times

        self._frequencies = frequencies
        self._spectrum_unit = spectrum_unit
        self._start_datetime = (
            np.datetime64(start_datetime) if start_datetime is not None else None
        )

        # by default, the background is evaluated over the whole spectrogram
        self._start_background_index = 0
        self._end_background_index = self.num_times
        # the background interval can be set after instantiation
        self._start_background: typing.Optional[str] = None
        self._end_background: typing.Optional[str] = None

        # finally check that the spectrogram arrays are matching in shape
        self._check_shapes()

    @property
    def dynamic_spectra(self) -> npt.NDArray[np.float32]:
        """The dynamic spectra array.

        Returns the 2D array of spectrogram data with shape (num_frequencies, num_times),
        where values are in the units specified by `spectrum_unit`.
        """
        return self._dynamic_spectra

    @property
    def times(self) -> npt.NDArray[np.float32]:
        """A 1D array representing the elapsed time of each spectrum, in seconds, relative to the first
        in the spectrogram."""
        return self._times

    @property
    def num_times(self) -> int:
        """The size of the times array. Equivalent to the number of spectrums in the spectrogram."""
        return len(self._times)

    @property
    def time_resolution(self) -> float:
        """The time resolution of the dynamic spectra.

        Represents the spacing between consecutive time values in the times array,
        calculated as the median difference between adjacent elements.
        """
        return compute_resolution(self._times)

    @property
    def time_range(self) -> float:
        """The time range of the dynamic spectra.

        Represents the difference between the first and last time values
        in the times array.
        """
        return compute_range(self._times)

    @property
    def frequencies(self) -> npt.NDArray[np.float32]:
        """A 1D array representing the physical frequencies assigned to each spectral component, in Hz."""
        return self._frequencies

    @property
    def num_frequencies(self) -> int:
        """The number of spectral components in the spectrogram.

        Equivalent to the size of the frequencies array.
        """
        return len(self._frequencies)

    @property
    def frequency_resolution(self) -> float:
        """The frequency resolution of the dynamic spectra.

        Represents the spacing between consecutive frequency values in the frequencies array,
        calculated as the median difference between adjacent elements.
        """
        return compute_resolution(self._frequencies)

    @property
    def frequency_range(self) -> float:
        """The frequency range covered by the dynamic spectra.

        Represents the difference between the minimum and maximum frequency
        values in the frequencies array.
        """
        return compute_range(self._frequencies)

    @property
    def start_datetime_is_set(self) -> bool:
        """Indicates whether the start datetime for the spectrogram has been set."""
        return self._start_datetime is not None

    @property
    def start_datetime(self) -> np.datetime64:
        """The datetime assigned to the first spectrum in `dynamic_spectra`.

        :raises AttributeError: If the start_datetime has not been set.
        """
        if self._start_datetime is None:
            raise ValueError(f"A start time has not been set.")
        return self._start_datetime

    @property
    def datetimes(self) -> npt.NDArray[np.datetime64]:
        """The datetimes associated with each spectrum in `dynamic_spectra`.

        Returns a list of datetime objects, calculated by adding the elapsed
        times in the times array to the start_datetime.
        """
        return self.start_datetime + (1e6 * self._times).astype("timedelta64[us]")

    @property
    def spectrum_unit(self) -> SpectrumUnit:
        """The units associated with the `dynamic_spectra` array."""
        return self._spectrum_unit

    @property
    def start_background(self) -> typing.Optional[str]:
        """The start of the background interval.

        Returns a string-formatted datetime up to seconds precision, or None
        if the background interval has not been set.
        """
        return self._start_background

    @property
    def end_background(self) -> typing.Optional[str]:
        """The end of the background interval.

        Returns a string-formatted datetime up to seconds precision, or None
        if the background interval has not been set.
        """
        return self._end_background

    def compute_background_spectrum(self) -> npt.NDArray[np.float32]:
        """Compute the background spectrum by averaging the dynamic spectra in time.

        By default, the entire dynamic spectra is averaged. Use set_background to
        specify a custom background interval.

        :return: A 1D array representing the time-averaged dynamic spectra over the
        specified background interval.
        """
        return np.nanmean(
            self._dynamic_spectra[
                :, self._start_background_index : self._end_background_index + 1
            ],
            axis=-1,
        )

    def compute_dynamic_spectra_dBb(self) -> npt.NDArray[np.float32]:
        """Compute the dynamic spectra in units of decibels above the background spectrum.

        The computation applies logarithmic scaling based on the `spectrum_unit`.

        :raises NotImplementedError: If the spectrum_unit is unrecognised.
        :return: A 2D array with the same shape as `dynamic_spectra`, representing
        the values in decibels above the background.
        """
        # Create a 'background' spectrogram where each spectrum is identically the background spectrum
        background_spectrum = self.compute_background_spectrum()
        background_spectra = background_spectrum[:, np.newaxis]
        # Suppress divide by zero and invalid value warnings for this block of code
        with np.errstate(divide="ignore", invalid="ignore"):
            if (
                self._spectrum_unit == SpectrumUnit.AMPLITUDE
                or self._spectrum_unit == SpectrumUnit.DIGITS
            ):
                dynamic_spectra_dBb = 10 * np.log10(
                    self._dynamic_spectra / background_spectra
                )
            elif self._spectrum_unit == SpectrumUnit.POWER:
                dynamic_spectra_dBb = 20 * np.log10(
                    self._dynamic_spectra / background_spectra
                )
            else:
                raise NotImplementedError(
                    f"{self._spectrum_unit} is unrecognised; decibel conversion is uncertain!"
                )
        return dynamic_spectra_dBb.astype(np.float32)

    def format_start_time(self) -> str:
        """Format the datetime assigned to the first spectrum in the dynamic spectra.

        :return: A string representation of the `start_datetime`, up to seconds precision.
        """
        dt = self.start_datetime.astype(datetime.datetime)
        return datetime.datetime.strftime(dt, spectre_core.config.TimeFormat.DATETIME)

    def set_background(self, start_background: str, end_background: str) -> None:
        """Set the background interval for computing the background spectrum, and doing
        background subtractions.

        :param start_background: The start time of the background interval, formatted as
        a string in the format `TimeFormat.DATETIME` (up to seconds precision).
        :param end_background: The end time of the background interval, formatted as
        a string in the format `TimeFormat.DATETIME` (up to seconds precision).
        """
        self._start_background = start_background
        self._end_background = end_background
        self._update_background_indices_from_interval(
            self._start_background, self._end_background
        )

    def _update_background_indices_from_interval(
        self, start_background: str, end_background: str
    ) -> None:
        start_background_datetime = np.datetime64(
            datetime.datetime.strptime(
                start_background, spectre_core.config.TimeFormat.DATETIME
            )
        )
        end_background_datetime = np.datetime64(
            datetime.datetime.strptime(
                end_background, spectre_core.config.TimeFormat.DATETIME
            )
        )
        self._start_background_index = find_closest_index(
            start_background_datetime, self.datetimes, enforce_strict_bounds=True
        )
        self._end_background_index = find_closest_index(
            end_background_datetime, self.datetimes, enforce_strict_bounds=True
        )

    def _check_shapes(self) -> None:
        """Check that the data array shapes are consistent with one another."""
        num_spectrogram_dims = np.ndim(self._dynamic_spectra)
        # Check if 'dynamic_spectra' is a 2D array
        if num_spectrogram_dims != 2:
            raise ValueError(
                f"Expected dynamic spectrogram to be a 2D array, but got {num_spectrogram_dims}D array"
            )
        dynamic_spectra_shape = self.dynamic_spectra.shape
        # Check if the dimensions of 'dynamic_spectra' are consistent with the time and frequency arrays
        if dynamic_spectra_shape[0] != self.num_frequencies:
            raise ValueError(
                f"Mismatch in number of frequency bins: Expected {self.num_frequencies}, but got {dynamic_spectra_shape[0]}"
            )

        if dynamic_spectra_shape[1] != self.num_times:
            raise ValueError(
                f"Mismatch in number of time bins: Expected {self.num_times}, but got {dynamic_spectra_shape[1]}"
            )

    def integrate_over_frequency(
        self, correct_background: bool = False, peak_normalise: bool = False
    ) -> npt.NDArray[np.float32]:
        """Numerically integrate the spectrogram over the frequency axis.

        :param correct_background: Indicates whether to subtract the background after
        computing the integral, defaults to False
        :param peak_normalise: Indicates whether to normalise the integral such that
        the peak value is equal to unity, defaults to False
        :return: A 1D array containing each spectrum numerically integrated over the
        frequency axis.
        """
        # numerically integrate over frequency
        I = np.trapz(self._dynamic_spectra, self._frequencies, axis=0)

        if correct_background:
            I = subtract_background(
                I, self._start_background_index, self._end_background_index
            )
        if peak_normalise:
            I = normalise_peak_intensity(I)
        return I

    def get_frequency_cut(
        self, at_time: float | str, dBb: bool = False, peak_normalise: bool = False
    ) -> FrequencyCut:
        """Retrieve a cut of the dynamic spectra at a specific time.

        If `at_time` does not match exactly with a time in `times`, the closest match
        is selected. The cut represents one of the spectrums in the spectrogram.

        :param at_time: The requested time for the cut. If a string, it is parsed
        as a datetime. If a float, it is treated as elapsed time since the first spectrum.
        :param dBb: If True, returns the cut in decibels above the background,
        defaults to False.
        :param peak_normalise: If True, normalises the cut such that its peak value
        is equal to 1. Ignored if dBb is True, defaults to False.
        :raises ValueError: If at_time is not a recognised type.
        :return: A FrequencyCut object containing the spectral values and associated metadata.
        """

        if isinstance(at_time, str):
            _at_datetime = np.datetime64(
                datetime.datetime.strptime(
                    at_time, spectre_core.config.TimeFormat.DATETIME
                )
            )
            index_of_cut = find_closest_index(
                _at_datetime, self.datetimes, enforce_strict_bounds=True
            )
            time_of_cut = self.datetimes[index_of_cut]

        elif isinstance(at_time, float):
            _at_time = np.float32(at_time)
            index_of_cut = find_closest_index(
                _at_time, self._times, enforce_strict_bounds=True
            )
            time_of_cut = self.times[index_of_cut]

        else:
            raise ValueError(f"'at_time' type '{type(at_time)}' is unsupported.")

        if dBb:
            ds = self.compute_dynamic_spectra_dBb()
        else:
            ds = self._dynamic_spectra

        cut = ds[
            :, index_of_cut
        ].copy()  # make a copy so to preserve the spectrum on transformations of the cut

        if dBb:
            if peak_normalise:
                warnings.warn(
                    "Ignoring frequency cut normalisation, since dBb units have been specified"
                )
        else:
            if peak_normalise:
                cut = normalise_peak_intensity(cut)

        return FrequencyCut(time_of_cut, self._frequencies, cut, self._spectrum_unit)

    def get_time_cut(
        self,
        at_frequency: float,
        dBb: bool = False,
        peak_normalise=False,
        correct_background=False,
        return_time_type: TimeType = TimeType.RELATIVE,
    ) -> TimeCut:
        """Retrieve a cut of the dynamic spectra at a specific frequency.

        If `at_frequency` does not exactly match a frequency in `frequencies`, the
        closest match is selected. The cut represents the time series of some spectral
        component.

        :param at_frequency: The requested frequency for the cut, in Hz.
        :param dBb: If True, returns the cut in decibels above the background.
        Defaults to False.
        :param peak_normalise: If True, normalises the cut so its peak value is 1.
        Ignored if dBb is True. Defaults to False.
        :param correct_background: If True, subtracts the background from the cut.
        Ignored if dBb is True. Defaults to False.
        :param return_time_type: Specifies the type of time values in the cut
        (TimeType.RELATIVE or TimeType.DATETIMES). Defaults to TimeType.RELATIVE.
        :raises ValueError: If return_time_type is not recognised.
        :return: A TimeCut object containing the temporal values and associated metadata.
        """
        _at_frequency = np.float32(at_frequency)
        index_of_cut = find_closest_index(
            _at_frequency, self._frequencies, enforce_strict_bounds=True
        )
        frequency_of_cut = float(self.frequencies[index_of_cut])

        # dependent on the requested cut type, we return the dynamic spectra in the preferred units
        if dBb:
            ds = self.compute_dynamic_spectra_dBb()
        else:
            ds = self.dynamic_spectra

        cut = ds[
            index_of_cut, :
        ].copy()  # make a copy so to preserve the spectrum on transformations of the cut

        # Warn if dBb is used with background correction or peak normalisation
        if dBb:
            if correct_background or peak_normalise:
                warnings.warn(
                    "Ignoring time cut normalisation, since dBb units have been specified"
                )
        else:
            # Apply background correction if required
            if correct_background:
                cut = subtract_background(
                    cut, self._start_background_index, self._end_background_index
                )

            # Apply peak normalisation if required
            if peak_normalise:
                cut = normalise_peak_intensity(cut)

        if return_time_type == TimeType.DATETIMES:
            return TimeCut(frequency_of_cut, self.datetimes, cut, self.spectrum_unit)
        elif return_time_type == TimeType.RELATIVE:
            return TimeCut(frequency_of_cut, self.times, cut, self.spectrum_unit)
        else:
            raise ValueError(
                f"Invalid return_time_type. Got {return_time_type}, "
                f"expected one of 'datetimes' or 'seconds'"
            )

    def save(
        self,
        tag: str,
        origin: str,
        instrument: str,
        telescope: str,
        object: str,
        obs_alt: float,
        obs_lat: float,
        obs_lon: float,
        batches_dir_path: typing.Optional[str] = None,
    ) -> None:
        """Write the spectrogram and its associated metadata to a batch file in the FITS format."""
        # Create the primary HDU.
        primary_hdu = astropy.io.fits.PrimaryHDU(self.dynamic_spectra)

        primary_hdu.header.set("SIMPLE", True)
        primary_hdu.header.set("BITPIX", -32)
        primary_hdu.header.set("NAXIS", 2)
        primary_hdu.header.set("NAXIS1", self.num_times)
        primary_hdu.header.set("NAXIS2", self.num_frequencies)
        primary_hdu.header.set("EXTEND", True)

        comment = (
            "FITS (Flexible Image Transport System) format defined in Astronomy and Astrophysics Supplement Series v44/p363, v44/p371, v73/p359, v73/p365. Contact the NASA Science Office of Standards and Technology for the FITS Definition document #100 and other FITS information."
            ""
        )
        primary_hdu.header.add_comment(comment)

        start_datetime = typing.cast(
            datetime.datetime, self.datetimes[0].astype(datetime.datetime)
        )
        start_date = start_datetime.strftime("%Y-%m-%d")
        start_time = start_datetime.strftime("%H:%M:%S.%f")

        end_datetime = typing.cast(
            datetime.datetime, self.datetimes[-1].astype(datetime.datetime)
        )
        end_date = end_datetime.strftime("%Y-%m-%d")
        end_time = end_datetime.strftime("%H:%M:%S.%f")

        primary_hdu.header.set("DATE", start_date)
        primary_hdu.header.set("CONTENT", f"{start_date} dynamic spectrogram")
        primary_hdu.header.set("ORIGIN", f"{origin}")
        primary_hdu.header.set("TELESCOP", f"{telescope}")
        primary_hdu.header.set("INSTRUME", f"{instrument}")
        primary_hdu.header.set("OBJECT", f"{object}")

        primary_hdu.header.set("DATE-OBS", f"{start_date}")
        primary_hdu.header.set("TIME-OBS", f"{start_time}")
        primary_hdu.header.set("DATE-END", f"{end_date}")
        primary_hdu.header.set("TIME-END", f"{end_time}")

        primary_hdu.header.set("BZERO", 0)
        primary_hdu.header.set("BSCALE", 1)
        primary_hdu.header.set("BUNIT", f"{self.spectrum_unit.value}")

        primary_hdu.header.set("DATAMIN", np.nanmin(self.dynamic_spectra))
        primary_hdu.header.set("DATAMAX", np.nanmax(self.dynamic_spectra))

        primary_hdu.header.set("CRVAL1", f"{_seconds_of_day(start_datetime)}")
        primary_hdu.header.set("CRPIX1", 0)
        primary_hdu.header.set("CTYPE1", "TIME [UT]")
        primary_hdu.header.set("CDELT1", self.time_resolution)

        primary_hdu.header.set("CRVAL2", 0)
        primary_hdu.header.set("CRPIX2", 0)
        primary_hdu.header.set("CTYPE2", "Frequency [MHz]")
        primary_hdu.header.set("CDELT2", self.frequency_resolution)

        primary_hdu.header.set("OBS_LAT", f"{obs_lat}")
        primary_hdu.header.set("OBS_LAC", "N")
        primary_hdu.header.set("OBS_LON", f"{obs_lon}")
        primary_hdu.header.set("OBS_LOC", "W")
        primary_hdu.header.set("OBS_ALT", f"{obs_alt}")

        # Create the Binary table HDU, wrapping the arrays to mimic the e-CALLISTO FITS files.
        times = np.array([self.times])
        frequencies_MHz = self.frequencies * 1e-6  # Convert to MHz
        frequencies = np.array([frequencies_MHz])
        col1 = astropy.io.fits.Column(name="TIME", format="PD", array=times)
        col2 = astropy.io.fits.Column(name="FREQUENCY", format="PD", array=frequencies)
        cols = astropy.io.fits.ColDefs([col1, col2])
        bin_table_hdu = astropy.io.fits.BinTableHDU.from_columns(cols)
        bin_table_hdu.header.set("PCOUNT", 0)
        bin_table_hdu.header.set("GCOUNT", 1)
        bin_table_hdu.header.set("TFIELDS", 2)
        bin_table_hdu.header.set("TTYPE1", "TIME")
        bin_table_hdu.header.set("TFORM1", "D")
        bin_table_hdu.header.set("TTYPE2", "FREQUENCY")
        bin_table_hdu.header.set("TFORM2", "D")
        bin_table_hdu.header.set("TSCAL1", 1, "")
        bin_table_hdu.header.set("TZERO1", 0, "")
        bin_table_hdu.header.set("TSCAL2", 1, "")
        bin_table_hdu.header.set("TZERO2", 0, "")

        # Combine the HDUs, and write them to the filesystem as a file in the FITS format.
        hdul = astropy.io.fits.HDUList([primary_hdu, bin_table_hdu])

        dt = typing.cast(
            datetime.datetime, self.start_datetime.astype(datetime.datetime)
        )
        batch_parent_path = (
            batches_dir_path
            or spectre_core.config.paths.get_batches_dir_path(
                year=dt.year, month=dt.month, day=dt.day
            )
        )
        if not os.path.exists(batch_parent_path):
            os.makedirs(batch_parent_path)
        file_path = os.path.join(
            batch_parent_path, f"{self.format_start_time()}_{tag}.fits"
        )
        hdul.writeto(file_path, overwrite=True)


def _seconds_of_day(dt: datetime.datetime) -> float:
    start_of_day = datetime.datetime(dt.year, dt.month, dt.day)
    return (dt - start_of_day).total_seconds()



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing

import spectre_core.jobs
import spectre_core.logs

from ._factory import get_receiver
from ._config import Config


def _make_flowgraph_worker(
    config: Config,
    skip_validation: bool,
    spectre_data_dir_path: typing.Optional[str],
) -> spectre_core.jobs.Worker:
    receiver = get_receiver(config.receiver_name, config.receiver_mode)
    return spectre_core.jobs.make_worker(
        "flowgraph",
        receiver.activate_flowgraph,
        (config.tag, config.parameters, skip_validation),
        spectre_data_dir_path=spectre_data_dir_path,
    )


def _make_post_processing_worker(
    config: Config,
    skip_validation: bool,
    spectre_data_dir_path: typing.Optional[str],
) -> spectre_core.jobs.Worker:
    receiver = get_receiver(config.receiver_name, config.receiver_mode)
    return spectre_core.jobs.make_worker(
        "flowgraph",
        receiver.activate_post_processing,
        (config.tag, config.parameters, skip_validation),
        spectre_data_dir_path=spectre_data_dir_path,
    )


@spectre_core.logs.log_call
def record_signal(
    configs: list[Config],
    duration: float = 60,
    force_restart: bool = False,
    max_restarts: int = 5,
    skip_validation: bool = False,
    spectre_data_dir_path: typing.Optional[str] = None,
) -> int:
    """Capture data from SDRs in real time.

    :param config: A list of configs.
    :param duration: How long to record for, in seconds.
    :param force_restart: If specified, restart the recording if it fails at runtime.
    :param max_restarts: Maximum number of times the recording can be restarted before giving up.
    Only applies when force_restart is True. Defaults to 5.
    :param skip_validation: If True, skip validating the config parameters against the model.
    :return: 0 exit code on success.
    """
    flowgraph_workers = [
        _make_flowgraph_worker(config, skip_validation, spectre_data_dir_path)
        for config in configs
    ]
    spectre_core.jobs.start_job(
        flowgraph_workers, duration, force_restart, max_restarts
    )

    return 0


@spectre_core.logs.log_call
def record_spectrograms(
    configs: list[Config],
    duration: float = 60,
    force_restart: bool = False,
    max_restarts: int = 5,
    skip_validation: bool = False,
    spectre_data_dir_path: typing.Optional[str] = None,
) -> int:
    """Capture data from SDRs and post-process it into spectrograms in real time.

    :param configs: A list of configs.
    :param duration: How long to record for, in seconds.
    :param force_restart: If specified, restart the recording if it fails at runtime.
    :param max_restarts: Maximum number of times the recording can be restarted before giving up.
    Only applies when force_restart is True. Defaults to 5.
    :param skip_validation: If True, skip validating the config parameters against the model.
    :return: 0 exit code on success.
    """
    flowgraph_workers = [
        _make_flowgraph_worker(config, skip_validation, spectre_data_dir_path)
        for config in configs
    ]
    post_processing_workers = [
        _make_post_processing_worker(config, skip_validation, spectre_data_dir_path)
        for config in configs
    ]
    spectre_core.jobs.start_job(
        post_processing_workers + flowgraph_workers,
        duration=duration,
        force_restart=force_restart,
        max_restarts=max_restarts,
    )

    return 0



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing
import math

import pydantic


def skip_validator(info: pydantic.ValidationInfo) -> bool:
    return True if info.context and info.context.get("skip", False) else False


T = typing.TypeVar("T")


def validate_one_of(
    value: T,
    options: list[T],
    name: str = "Value",
):
    """
    Validate that the input value is one of a list of pre-defined options.

    :param value: The value to validate.
    :param options: The list of allowed options.
    :param name: The name of the value being validated (for error messages).
    :raises ValueError: If the value is not one of the options.
    """
    if value not in options:
        raise ValueError(f"{name} must be one of {options}. Got {value}.")


def validate_in_range(
    value: float | int,
    lower_bound: typing.Optional[float | int] = None,
    upper_bound: typing.Optional[float | int] | None = None,
    strict_lower: bool = False,
    strict_upper: bool = False,
    name: str = "Value",
) -> None:
    """
    Validate that a numeric value is within a specified interval.

    :param value: The value to validate.
    :param lower_bound: The value must be greater than `lower_bound`. Inclusive if `strict_lower` is False.
    :param upper_bound: The value must be less than `upper_bound`. Inclusive if `strict_upper` is False.
    :param strict_lower: If true, the value must be strictly greater than `lower_bound`.
    :param strict_upper: If true, the value must be strictly less than `upper_bound`.
    :param name: Include the name of the value being validated in the error message.
    :raises ValueError: If the value is outside the specified interval.
    """
    if lower_bound is not None:
        if strict_lower and value <= lower_bound:
            raise ValueError(
                f"{name} must be strictly greater than {lower_bound}. Got {value}."
            )
        if not strict_lower and value < lower_bound:
            raise ValueError(
                f"{name} must be greater than or equal to {lower_bound}. Got {value}."
            )

    if upper_bound is not None:
        if strict_upper and value >= upper_bound:
            raise ValueError(
                f"{name} must be strictly less than {upper_bound}. Got {value}."
            )
        if not strict_upper and value > upper_bound:
            raise ValueError(
                f"{name} must be less than or equal to {upper_bound}. Got {value}."
            )


def validate_window_size(window_size: int):
    """Check that the window size is a power of two."""
    if window_size & (window_size - 1) != 0:
        raise ValueError("The window size must be a power of 2")


def validate_nyquist_criterion(sample_rate: float, bandwidth: float) -> None:
    """Ensure that the Nyquist criterion is satisfied."""
    if sample_rate < bandwidth:
        raise ValueError(
            (
                f"Nyquist criterion has not been satisfied. "
                f"Sample rate must be greater than or equal to the bandwidth. "
                f"Got sample rate {sample_rate} [Hz], and bandwidth {bandwidth} [Hz]"
            )
        )


def validate_non_overlapping_steps(frequency_step: float, sample_rate: float) -> None:
    """Ensure that the stepped spectrograms are non-overlapping in the frequency domain."""

    if frequency_step < sample_rate:
        raise NotImplementedError(
            f"Spectre does not yet support spectral steps overlapping in frequency. "
            f"Got frequency step {frequency_step * 1e-6} [MHz] which is less than the sample "
            f"rate {sample_rate * 1e-6} [MHz]"
        )


def validate_num_steps_per_sweep(
    min_frequency: float, max_frequency: float, frequency_step: float
) -> None:
    """Ensure that there are at least two steps in frequency per sweep."""

    num_steps_per_sweep = math.floor((max_frequency - min_frequency) / frequency_step)
    if num_steps_per_sweep <= 1:
        raise ValueError(
            (
                f"We need strictly greater than one step per sweep. "
                f"Computed {num_steps_per_sweep} step per sweep"
            )
        )


def validate_num_samples_per_step(window_size: int, samples_per_step: int) -> None:
    """Ensure that the number of samples per step is greater than the window size."""
    if window_size >= samples_per_step:
        raise ValueError(
            (
                f"Window size must be strictly less than the number of samples per step. "
                f"Got window size {window_size}, which is more than or equal "
                f"to the number of samples per step {samples_per_step}"
            )
        )



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


"""Manage program logging."""

from ._process_types import ProcessType
from ._decorators import log_call
from ._configure import configure_root_logger, get_root_logger_state
from ._logs import Log, Logs, parse_log_file_name


__all__ = [
    "log_call",
    "configure_root_logger",
    "Log",
    "Logs",
    "ProcessType",
    "get_root_logger_state",
    "parse_log_file_name",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""IO operations on batched data files."""

from ._base import Base, BatchFile, parse_batch_file_name, parse_batch_file_path
from ._batches import Batches
from ._callisto import CallistoBatch
from ._iq_stream import IQMetadata, IQStreamBatch, IQStreamBatchExtension

__all__ = [
    "Base",
    "BatchFile",
    "parse_batch_file_name",
    "parse_batch_file_path",
    "Batches",
    "CallistoBatch",
    "IQMetadata",
    "IQStreamBatch",
    "IQStreamBatchExtension",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import time

from ._workers import Worker
from ._duration import Duration

_LOGGER = logging.getLogger(__name__)


class Job:
    def __init__(self, workers: list[Worker]) -> None:
        """Represents a collection of workers that run long-running tasks as
        multiprocessing processes.

        :param workers: A list of `Worker` instances to manage as part of the job.
        """
        self._workers = workers

    @property
    def workers_are_alive(self) -> bool:
        """Returns True if all managed workers are alive, and False otherwise."""
        return all([worker.is_alive for worker in self._workers])

    def start(
        self,
    ) -> None:
        """Tell each worker to call their functions in the background as multiprocessing processes."""
        if self.workers_are_alive:
            raise RuntimeError("A job cannot be started twice.")
        for worker in self._workers:
            worker.start()

    def kill(
        self,
    ) -> None:
        """Tell each worker to kill their processes, if the processes are still running."""
        _LOGGER.info("Killing workers...")
        for worker in self._workers:
            if worker.is_alive:
                worker.kill()

        _LOGGER.info("All workers successfully killed")

    def restart(
        self,
    ) -> None:
        """Tell each worker to restart its process."""
        for worker in self._workers:
            worker.restart()

    def monitor(
        self, duration: float, force_restart: bool = False, max_restarts: int = 5
    ) -> None:
        """
        Monitor the workers during execution and handle unexpected exits.

        Periodically checks worker processes within the specified runtime duration.
        If a worker exits unexpectedly:
        - Restarts all workers if `force_restart` is True.
        - Kills all workers and raises an exception if `force_restart` is False.

        :param duration: Total time to monitor the workers, in seconds.
        :param force_restart: Whether to restart all workers if one dies unexpectedly.
        :param max_restarts: Maximum number of times workers can be restarted before giving up and killing all workers.
        Only applies when force_restart is True. Defaults to 5.
        :raises RuntimeError: If a worker exits and `force_restart` is False.
        """
        _LOGGER.info("Monitoring workers...")
        start_time = time.time()

        restarts_remaining = max_restarts
        try:
            # Check that the elapsed time since the job started is within the total runtime configured by the user.
            while time.time() - start_time < duration:
                for worker in self._workers:
                    if not worker.is_alive:
                        error_message = (
                            f"Worker with name `{worker.name}` unexpectedly exited."
                        )
                        _LOGGER.error(error_message)
                        if force_restart:
                            if restarts_remaining > 0:
                                _LOGGER.info(
                                    f"Attempting restart ({restarts_remaining} restarts remaining)..."
                                )
                                restarts_remaining -= 1
                                self.restart()

                            else:
                                error_message = (
                                    f"Maximum number of restarts has been reached: {max_restarts}. "
                                    f"Killing all workers."
                                )
                                self.kill()
                                raise RuntimeError(error_message)
                        else:
                            self.kill()
                            raise RuntimeError(error_message)
                time.sleep(Duration.ONE_DECISECOND)  # Poll every 0.1 seconds

            # If the jobs total runtime has elapsed, kill all the workers
            _LOGGER.info("Job complete. Killing workers... ")
            self.kill()

        except KeyboardInterrupt:
            _LOGGER.info("Keyboard interrupt detected. Killing workers...")
            self.kill()


def start_job(
    workers: list[Worker],
    duration: float,
    force_restart: bool = False,
    max_restarts: int = 5,
) -> None:
    """Create and run a job with the specified workers.

    Starts the workers, monitors them for the specified runtime, and handles
    unexpected exits according to the `force_restart` policy.

    :param workers: A list of `Worker` instances to include in the job.
    :param duration: Total time to monitor the workers, in seconds.
    :param force_restart: Whether to restart all workers if one dies unexpectedly.
    :param max_restarts: Maximum number of times workers can be restarted before giving up and killing all workers.
    Only applies when force_restart is True. Defaults to 5.
    """
    job = Job(workers)
    job.start()
    job.monitor(duration, force_restart, max_restarts)



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import dataclasses
import typing

import numpy as np
import numpy.typing as npt
import astropy.io

import spectre_core.exceptions
import spectre_core.config
import spectre_core.spectrograms

from ._base import Base, BatchFile


@dataclasses.dataclass(frozen=True)
class IQStreamBatchExtension:
    """Supported extensions for a `IQStreamBatch`.

    :ivar FITS: Corresponds to the `.fits` file extension.
    :ivar BIN: Corresponds to the `.bin` file extension.
    :ivar HDR: Corresponds to the `.hdr` file extension.
    """

    FITS: str = "fits"
    BIN: str = "bin"
    HDR: str = "hdr"


class _BinFile(BatchFile[npt.NDArray[np.complex64]]):
    """Stores complex IQ samples in the binary format, as produced by the `gr-spectre`
    OOT module block `batched_file_sink`.
    """

    def read(self) -> npt.NDArray[np.complex64]:
        """Reads the binary file and returns the stored complex IQ samples.

        :return: The raw 32-bit floats are interpreted as 64-bit complex IQ samples.
        """
        return np.fromfile(self.file_path, dtype=np.complex64)


@dataclasses.dataclass
class IQMetadata:
    """Stores metadata produced by the `gr-spectre` OOT module block `batched_file_sink`.

    :ivar millisecond_correction: The millisecond component of the batch start time.
    :ivar center_frequencies: Center frequencies for each IQ sample, if the stream was frequency tagged.
    None otherwise.
    :ivar num_samples: Number of samples collected at each center frequency, if the stream was frequency
    tagged. None otherwise.
    """

    millisecond_correction: int
    center_frequencies: typing.Optional[npt.NDArray[np.float32]] = None
    num_samples: typing.Optional[npt.NDArray[np.int32]] = None

    def is_frequency_tagged(self) -> bool:
        """Check if the IQ metadata contains frequency tagging information.

        :return: True if frequency tagging information is present; False otherwise.
        """
        return (self.center_frequencies is not None) and (self.num_samples is not None)


class _HdrFile(BatchFile[IQMetadata]):
    """Stores IQ sample metadata produced by the `gr-spectre` OOT module block `batched_file_sink`, used
    to help parse the corresponding `.bin` file.

    File Structure:
        - If frequency tagged:
            (`<millisecond_correction>`, `<freq_0>`, `<num_samples_0>`, `<freq_1>`, `<num_samples_1>`, ...)
            All values are stored as 32-bit floats.
            - The first value is the millisecond component for the batch start time.
            - Subsequent tuples (`<freq_n>`, `<num_samples_n>`) indicate that `<num_samples_n>` samples were collected at `<freq_n>`.
        - If not frequency tagged:
            (`<millisecond_correction>`)
            Only the millisecond correction is present, with no frequency information.

    This format enables mapping IQ samples in the binary file to their corresponding center frequencies, if applicable.
    """

    def read(self) -> IQMetadata:
        """Parses the binary contents of the `.hdr` file to extract IQ sample metadata.

        :return: An instance of `IQMetadata` containing the parsed metadata, including the millisecond correction
        and, if applicable, frequency tag details.
        """
        hdr_contents = np.fromfile(self.file_path, dtype=np.float32)

        millisecond_correction_f = float(hdr_contents[0])
        if not millisecond_correction_f.is_integer():
            raise TypeError(
                f"Expected integer value for millisecond correction, but got {millisecond_correction_f}"
            )
        millisecond_correction = int(millisecond_correction_f)

        if hdr_contents.size == 1:
            return IQMetadata(millisecond_correction)
        else:
            # Center frequencies are stored at every second entry, starting from the first index.
            center_frequencies = hdr_contents[1::2]

            # Sample counts are located at every second entry, starting from the second index.
            # The values are stored as 32-bit floats but are interpreted as integers.
            num_samples_f = hdr_contents[2::2]
            if not all(num_samples_f == num_samples_f.astype(int)):
                raise spectre_core.exceptions.InvalidSweepMetadataError(
                    "Number of samples per frequency is expected to describe an integer"
                )
            num_samples = num_samples_f.astype(np.int32)

            if len(center_frequencies) != len(num_samples):
                raise spectre_core.exceptions.InvalidSweepMetadataError(
                    "Center frequencies and number of samples arrays are not the same length"
                )

            return IQMetadata(millisecond_correction, center_frequencies, num_samples)


class _FitsFile(BatchFile[spectre_core.spectrograms.Spectrogram]):
    """Stores spectrogram data in the FITS file format, as generated by Spectre from a stream of IQ samples."""

    def read(self) -> spectre_core.spectrograms.Spectrogram:
        """Read the FITS file and create a spectrogram.

        :return: A `Spectrogram` containing the parsed FITS file data.
        """
        with astropy.io.fits.open(self.file_path, mode="readonly") as hdulist:
            primary_hdu = hdulist["PRIMARY"]
            dynamic_spectra = primary_hdu.data
            bunit = primary_hdu.header["BUNIT"]

            date_obs = primary_hdu.header["DATE-OBS"]
            time_obs = primary_hdu.header["TIME-OBS"]
            spectrogram_start_datetime = datetime.datetime.strptime(
                f"{date_obs}T{time_obs}",
                spectre_core.config.TimeFormat.PRECISE_DATETIME,
            )

            bintable_hdu = hdulist[1]
            times = bintable_hdu.data["TIME"][0]
            frequencies = bintable_hdu.data["FREQUENCY"][0] * 1e6  # Convert to Hz

        # bunit is interpreted as a SpectrumUnit.
        spectrum_unit = spectre_core.spectrograms.SpectrumUnit(bunit)
        return spectre_core.spectrograms.Spectrogram(
            dynamic_spectra,
            times,
            frequencies,
            spectrum_unit,
            spectrogram_start_datetime,
        )


class IQStreamBatch(Base):

    def __init__(self, batches_dir_path: str, start_time: str, tag: str) -> None:
        """A batch of data derived from a stream of IQ samples from some receiver.

        Supports the following extensions:
        - `.fits`
        - `.bin`
        - `.hdr`


        :param start_time: The start time of the batch.
        :param tag: The batch name tag.
        """
        super().__init__(batches_dir_path, start_time, tag)

        self.add_file(_FitsFile, IQStreamBatchExtension.FITS)
        self.add_file(_BinFile, IQStreamBatchExtension.BIN)
        self.add_file(_HdrFile, IQStreamBatchExtension.HDR)

    @property
    def fits_file(self) -> _FitsFile:
        """The batch file corresponding to the `.fits` extension."""
        return typing.cast(_FitsFile, self.get_file(IQStreamBatchExtension.FITS))

    @property
    def bin_file(self) -> _BinFile:
        """The batch file corresponding to the `.bin` extension."""
        return typing.cast(_BinFile, self.get_file(IQStreamBatchExtension.BIN))

    @property
    def hdr_file(self) -> _HdrFile:
        """The batch file corresponding to the `.hdr` extension."""
        return typing.cast(_HdrFile, self.get_file(IQStreamBatchExtension.HDR))

    @property
    def spectrogram_file(self) -> _FitsFile:
        return self.fits_file



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses


@dataclasses.dataclass
class PanelFormat:
    """Specifies formatting options for a panel, including font sizes, line styles,
    colour maps, and general visual settings.

    These formatting values can be applied consistently across all panels within a `PanelStack`,
    but are optional.

    :ivar small_size: Font size for small text elements, defaults to 18.
    :ivar medium_size: Font size for medium text elements, defaults to 21.
    :ivar large_size: Font size for large text elements, defaults to 24.
    :ivar line_width: Thickness of lines in the plot, defaults to 3.
    :ivar line_color: Colour used for line elements, defaults to "lime".
    :ivar line_cmap: Colormap applied to line-based visual elements, defaults to "winter".
    :ivar style: Matplotlib style applied to the panel, defaults to "dark_background".
    :ivar spectrogram_cmap: Colormap applied to spectrogram plots, defaults to "gnuplot2".
    """

    small_size: int = 18
    medium_size: int = 21
    large_size: int = 24
    line_width: int = 3
    line_color: str = "lime"
    line_cmap: str = "winter"
    style: str = "dark_background"
    spectrogram_cmap: str = "gnuplot2"



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Spectre custom exceptions.
"""

import warnings
from functools import wraps
from typing import TypeVar, Callable, Any, cast

F = TypeVar("F", bound=Callable[..., Any])


def deprecated(message: str) -> Callable[[F], F]:
    """A decorator to mark functions as deprecated.

    :param message: Warning message explaining what to use instead
    """

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            warnings.warn(
                f"{func.__name__} is deprecated. {message}",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return cast(F, wrapper)

    return decorator


class BatchNotFoundError(KeyError): ...


class ModeNotFoundError(KeyError): ...


class ReceiverNotFoundError(KeyError): ...


class InvalidTagError(ValueError): ...


class InvalidSweepMetadataError(ValueError): ...



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import datetime
import typing

import numpy as np

import spectre_core.batches
import spectre_core.spectrograms
import spectre_core.fields

from ._base import Base, BaseModel
from ._stfft import (
    get_buffer,
    get_window,
    get_times,
    get_num_spectrums,
    get_frequencies,
    get_fftw_obj,
    stfft,
    WindowType,
)

_LOGGER = logging.getLogger(__name__)


class FixedCenterFrequencyModel(BaseModel):
    window_size: spectre_core.fields.Field.window_size = 1024
    window_hop: spectre_core.fields.Field.window_hop = 1024
    window_type: spectre_core.fields.Field.window_type = "blackman"
    center_frequency: spectre_core.fields.Field.center_frequency = 95.8e6
    sample_rate: spectre_core.fields.Field.sample_rate = 1000000
    frequency_resolution: spectre_core.fields.Field.frequency_resolution = 0
    time_resolution: spectre_core.fields.Field.time_resolution = 0
    batch_size: spectre_core.fields.Field.batch_size = 3
    keep_signal: spectre_core.fields.Field.keep_signal = False


class FixedCenterFrequency(
    Base[FixedCenterFrequencyModel, spectre_core.batches.IQStreamBatch]
):
    def __init__(
        self,
        tag: str,
        model: FixedCenterFrequencyModel,
        batch_cls: typing.Type[spectre_core.batches.IQStreamBatch],
    ) -> None:
        super().__init__(tag, model, batch_cls)
        self.__model = model

        # Make the window.
        self.__window = get_window(
            WindowType(self.__model.window_type), self.__model.window_size
        )

        # Pre-allocate the buffer.
        self.__buffer = get_buffer(self.__model.window_size)

        # Defer the expensive FFTW plan creation until the first batch is being processed.
        # With this approach, we avoid a bug where filesystem events are missed because
        # the watchdog observer isn't set up in time before the receiver starts capturing data.
        self.__fftw_obj = None

    @property
    def _watch_extension(self) -> str:
        return spectre_core.batches.IQStreamBatchExtension.BIN

    def process(
        self, batch: spectre_core.batches.IQStreamBatch
    ) -> spectre_core.spectrograms.Spectrogram:
        """Compute the spectrogram of IQ samples captured at a fixed center frequency."""
        _LOGGER.info(f"Reading {batch.bin_file.file_name}")
        iq_data = batch.bin_file.read()

        _LOGGER.info(f"Reading {batch.hdr_file.file_name}")
        iq_metadata = batch.hdr_file.read()

        if self.__fftw_obj is None:
            _LOGGER.info(f"Creating the FFTW plan")
            self.__fftw_obj = get_fftw_obj(self.__buffer)

        _LOGGER.info("Executing the short-time FFT")
        dynamic_spectra = stfft(
            self.__fftw_obj,
            self.__buffer,
            iq_data,
            self.__window,
            self.__model.window_hop,
        )

        # Compute the physical times we'll assign to each spectrum.
        num_spectrums = get_num_spectrums(
            iq_data.size, self.__model.window_size, self.__model.window_hop
        )
        times = get_times(
            num_spectrums, self.__model.sample_rate, self.__model.window_hop
        )

        # Get the physical frequencies assigned to each spectral component, shift the zero frequency to the middle of the
        # spectrum, then translate the array up from the baseband.
        frequencies = (
            np.fft.fftshift(
                get_frequencies(self.__model.window_size, self.__model.sample_rate)
            )
            + self.__model.center_frequency
        )

        # Shift the zero-frequency component to the middle of the spectrum.
        dynamic_spectra = np.fft.fftshift(dynamic_spectra, axes=0)

        # Account for the millisecond correction.
        start_datetime = batch.start_datetime + datetime.timedelta(
            milliseconds=iq_metadata.millisecond_correction
        )

        _LOGGER.info("Creating the spectrogram")
        spectrogram = spectre_core.spectrograms.Spectrogram(
            dynamic_spectra,
            times,
            frequencies,
            spectre_core.spectrograms.SpectrumUnit.AMPLITUDE,
            start_datetime,
        )

        spectrogram = spectre_core.spectrograms.time_average(
            spectrogram, resolution=self.__model.time_resolution
        )
        spectrogram = spectre_core.spectrograms.frequency_average(
            spectrogram, resolution=self.__model.frequency_resolution
        )

        _LOGGER.info("Spectrogram created successfully")

        if not self.__model.keep_signal:
            _LOGGER.info(f"Deleting {batch.bin_file.file_path}")
            batch.bin_file.delete()

            _LOGGER.info(f"Deleting {batch.hdr_file.file_path}")
            batch.hdr_file.delete()

        return spectrogram



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""File system path definitions."""

import os
import pathlib
from typing import Optional, Dict


DEFAULT_SPECTRE_DATA_DIR_PATH = pathlib.Path(os.curdir) / ".spectre_data"


class Paths:
    def __init__(self, env: Optional[Dict[str, str]] = None):
        """Manages file system paths for Spectre.

        :param env: A dictionary representing environment variables. Defaults to `os.environ`.
        """
        self._env = env or os.environ
        self.__make_dirs()

    def get_spectre_data_dir_path(self) -> str:
        """Get the base directory for Spectre data.

        :return: The path stored in the `SPECTRE_DATA_DIR_PATH` environment variable, or the default.
        """
        return str(
            pathlib.Path(
                self._env.get("SPECTRE_DATA_DIR_PATH", DEFAULT_SPECTRE_DATA_DIR_PATH)
            ).absolute()
        )

    def set_spectre_data_dir_path(self, spectre_data_dir_path: str) -> None:
        """Set the `SPECTRE_DATA_DIR_PATH` environment variable and create necessary directories.

        :param spectre_data_dir_path: The new base directory for Spectre data.
        """
        self._env["SPECTRE_DATA_DIR_PATH"] = spectre_data_dir_path
        self.__make_dirs()

    def __make_dirs(self) -> None:
        self.__mkdir(pathlib.Path(self.get_spectre_data_dir_path()))
        self.__mkdir(pathlib.Path(self.get_batches_dir_path()))
        self.__mkdir(pathlib.Path(self.get_logs_dir_path()))
        self.__mkdir(pathlib.Path(self.get_configs_dir_path()))

    def get_batches_dir_path(
        self,
        year: Optional[int] = None,
        month: Optional[int] = None,
        day: Optional[int] = None,
    ) -> str:
        """Get the directory for batched data files, optionally with a date-based subdirectory."""
        return str(
            self.__get_date_based_dir_path(
                pathlib.Path(self.get_spectre_data_dir_path()) / "batches",
                year,
                month,
                day,
            )
        )

    def get_logs_dir_path(
        self,
        year: Optional[int] = None,
        month: Optional[int] = None,
        day: Optional[int] = None,
    ) -> str:
        """Get the directory for log files, optionally with a date-based subdirectory."""
        return str(
            self.__get_date_based_dir_path(
                pathlib.Path(self.get_spectre_data_dir_path()) / "logs",
                year,
                month,
                day,
            )
        )

    def get_configs_dir_path(self) -> str:
        """Get the directory for configuration files."""
        return str(pathlib.Path(self.get_spectre_data_dir_path()) / "configs")

    def __get_date_based_dir_path(
        self,
        base_dir: pathlib.Path,
        year: Optional[int],
        month: Optional[int],
        day: Optional[int],
    ) -> pathlib.Path:
        """Append a date-based directory onto the base directory."""
        if day and not (year and month):
            raise ValueError("A day requires both a month and a year")
        if month and not year:
            raise ValueError("A month requires a year")

        date_components = []
        if year:
            date_components.append(f"{year:04}")
        if month:
            date_components.append(f"{month:02}")
        if day:
            date_components.append(f"{day:02}")

        return base_dir.joinpath(*date_components)

    def __mkdir(self, path: pathlib.Path) -> None:
        """Create a directory if it doesn't already exist."""
        path.mkdir(parents=True, exist_ok=True)


paths = Paths()



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses

import spectre_core.events
import spectre_core.flowgraphs
import spectre_core.models
import spectre_core.batches

from ._register import register_receiver
from ._base import Base
from ._names import ReceiverName


@dataclasses.dataclass(frozen=True)
class _Mode:
    FIXED_CENTER_FREQUENCY = "fixed_center_frequency"
    SWEPT_CENTER_FREQUENCY = "swept_center_frequency"


@register_receiver(ReceiverName.RSPDUO)
class RSPduo(Base):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.add_mode(
            _Mode.FIXED_CENTER_FREQUENCY,
            spectre_core.models.RSPduoFixedCenterFrequency,
            spectre_core.flowgraphs.RSPduoFixedCenterFrequency,
            spectre_core.events.FixedCenterFrequency,
            spectre_core.batches.IQStreamBatch,
        )

        self.add_mode(
            _Mode.SWEPT_CENTER_FREQUENCY,
            spectre_core.models.RSPduoSweptCenterFrequency,
            spectre_core.flowgraphs.RSPduoSweptCenterFrequency,
            spectre_core.events.SweptCenterFrequency,
            spectre_core.batches.IQStreamBatch,
        )
